import numpy as np
from pydantic import BaseModel, ConfigDict

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier

from .. import PyAMLException
from ..common import abstract
from ..common.element import Element, ElementConfigModel
from ..control.deviceaccess import DeviceAccess

# Define the main class name for this module
PYAMLCLASS = "RFTransmitter"


class ConfigModel(ElementConfigModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    voltage: DeviceAccess | None = None
    """Device to apply cavity voltage"""
    phase: DeviceAccess | None = None
    """Device to apply cavity phase"""
    cavities: list[str]
    """List of cavity names connected to this transmitter"""
    harmonic: float = 1.0
    """Harmonic frequency ratio, 1.0 for main frequency"""
    distribution: float = 1.0
    """RF distribution (Part of the total RF voltage powered by this transmitter)"""


class RFTransmitter(Element):
    """
    Class that handle a RF transmitter
    """

    def __init__(self, cfg: ConfigModel):
        super().__init__(cfg.name)
        self._cfg = cfg
        self.__voltage = None
        self.__phase = None

    @property
    def voltage(self) -> abstract.ReadWriteFloatScalar:
        if self.__voltage is None:
            raise PyAMLException(
                f"{str(self)} is unattached or has no voltage device defined"
            )
        return self.__voltage

    @property
    def phase(self) -> abstract.ReadWriteFloatScalar:
        if self.__phase is None:
            raise PyAMLException(
                f"{str(self)} is unattached or has no phase device defined"
            )
        return self.__phase

    def attach(
        self,
        peer,
        voltage: abstract.ReadWriteFloatScalar,
        phase: abstract.ReadWriteFloatScalar,
    ) -> Self:
        # Attach voltage and phase attribute and returns a new reference
        obj = self.__class__(self._cfg)
        obj.__voltage = voltage
        obj.__phase = phase
        obj._peer = peer
        return obj



from abc import ABCMeta, abstractmethod

from numpy import array, double

# Float ----------------------------------------------------------------


class ReadFloatScalar(metaclass=ABCMeta):
    """
    Abstract class providing read access to a scalar double
    """

    @abstractmethod
    def get(self) -> double:
        """Get the value"""
        pass

    @abstractmethod
    def unit(self) -> str:
        """Get the unit of the value"""
        pass


class ReadWriteFloatScalar(ReadFloatScalar):
    """
    Abstract class providing read write access to a scalar double
    """

    @abstractmethod
    def set(self, value: double):
        """Set the value"""
        pass

    # Sets the value and wait that the read value reach the setpoint
    @abstractmethod
    def set_and_wait(self, value: double):
        """Set the value and wait that setpoint is reached"""
        pass


class ReadFloatArray(metaclass=ABCMeta):
    """
    Abstract class providing read access to a vector of float
    """

    @abstractmethod
    def get(self) -> array:
        """Get the value"""
        pass

    @abstractmethod
    def unit(self) -> list[str]:
        """Get the unit of the values"""
        pass


class ReadWriteFloatArray(ReadFloatScalar):
    """
    Abstract class providing read write access to a vector of double
    """

    @abstractmethod
    def set(self, value: array):
        """Set the values"""
        pass

    # Sets the value and waits that the read value reach the setpoint
    @abstractmethod
    def set_and_wait(self, value: array):
        """Set the values and wait that setpoints are reached"""
        pass


class RWMapper(ReadWriteFloatScalar):
    """
    Class mapping a scalar to an element of an array
    """

    def __init__(self, bind, idx: int):
        self.bind = bind
        self.idx = idx

    # Gets the value
    def get(self) -> float:
        return self.bind.get()[self.idx]

    # Sets the value
    def set(self, value: float):
        arr = self.bind.get()
        arr[self.idx] = value
        self.bind.set(arr)

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    # Return the unit
    def unit(self) -> str:
        return self.bind.unit()[self.idx]

    # Return the mapped index
    def index(self) -> int:
        return self.idx



"""
Array configuration
"""

from pydantic import BaseModel, ConfigDict

from pyaml.common.exception import PyAMLException

from ..common.element_holder import ElementHolder


class ArrayConfigModel(BaseModel):
    """
    Base class for array configuration

    Parameters
    ----------
    name : str
        Family name
    elements : list[str]
        List of pyaml element names
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    name: str
    elements: list[str]


class ArrayConfig(object):
    """
    Base class that implements configuration for access to arrays (families)
    """

    def __init__(self, cfg: ArrayConfigModel):
        self._cfg = cfg

    def fill_array(self, holder: ElementHolder):
        raise PyAMLException("Array.fill_array() is not subclassed")



import numpy as np
from pydantic import BaseModel, ConfigDict

from .. import PyAMLException
from ..common.element import __pyaml_repr__
from ..control.deviceaccess import DeviceAccess
from .model import MagnetModel

# Define the main class name for this module
PYAMLCLASS = "IdentityMagnetModel"


class ConfigModel(BaseModel):
    """
    Configuration model for identity magnet model

    Parameters
    ----------
    powerconverter : DeviceAccess, optional
        Power converter device to apply current
    physics : DeviceAccess, optional
        Magnet device to apply strength
    unit : str
        Unit of the strength (i.e. 1/m or m-1)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    powerconverter: DeviceAccess | None = None
    physics: DeviceAccess | None = None
    unit: str


class IdentityMagnetModel(MagnetModel):
    """
    Class that map value to underlying device without conversion
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        self.__unit = cfg.unit
        if cfg.physics is None and cfg.powerconverter is None:
            raise PyAMLException(
                "Invalid IdentityMagnetModel configuration,"
                "physics or powerconverter device required"
            )
        if cfg.physics is not None and cfg.powerconverter is not None:
            raise PyAMLException(
                "Invalid IdentityMagnetModel configuration,"
                "physics or powerconverter device required but not both"
            )
        if cfg.physics:
            self.__device = cfg.physics
        else:
            self.__device = cfg.powerconverter

    def compute_hardware_values(self, strengths: np.array) -> np.array:
        return strengths

    def compute_strengths(self, currents: np.array) -> np.array:
        return currents

    def get_strength_units(self) -> list[str]:
        return [self.__unit]

    def get_hardware_units(self) -> list[str]:
        return [self.__unit]

    def get_devices(self) -> list[DeviceAccess]:
        return [self.__device]

    def set_magnet_rigidity(self, brho: np.double):
        pass

    def has_physics(self) -> bool:
        return self._cfg.physics is not None

    def has_hardware(self) -> bool:
        return self._cfg.powerconverter is not None

    def __repr__(self):
        return __pyaml_repr__(self)



from abc import ABCMeta, abstractmethod
from typing import Union

import numpy.typing as npt

from .readback_value import Value


class DeviceAccess(metaclass=ABCMeta):
    """
    Abstract class providing access to a control system float variable
    """

    @abstractmethod
    def name(self) -> str:
        """Return the name of the variable"""
        pass

    @abstractmethod
    def measure_name(self) -> str:
        """Return the name of the measure"""
        pass

    @abstractmethod
    def set(self, value: float):
        """Write a control system device variable (i.e. a power supply current)"""
        pass

    @abstractmethod
    def set_and_wait(self, value: float):
        """Write a control system device variable (i.e. a power supply current)"""
        pass

    @abstractmethod
    def get(self) -> float:
        """Return the setpoint of a control system device variable"""
        pass

    @abstractmethod
    def readback(self) -> Union[Value, npt.NDArray[Value]]:
        """Return the measured variable"""
        pass

    @abstractmethod
    def unit(self) -> str:
        """Return the variable unit"""
        pass



from ..common import abstract
from ..common.constants import HORIZONTAL_KICK_SIGN
from ..lattice.polynom_info import PolynomInfo
from .corrector import RWCorrectorAngle
from .magnet import Magnet, MagnetConfigModel

# Define the main class name for this module
PYAMLCLASS = "HCorrector"


class ConfigModel(MagnetConfigModel): ...


class HCorrector(Magnet):
    """Horizontal Corrector class"""

    polynom = PolynomInfo("PolynomB", 0, HORIZONTAL_KICK_SIGN)

    def __init__(self, cfg: ConfigModel):
        super().__init__(
            cfg.name,
            cfg.model if hasattr(cfg, "model") else None,
        )
        self._cfg = cfg
        self.__angle = RWCorrectorAngle(self)

    @property
    def angle(self) -> abstract.ReadWriteFloatScalar:
        """
        Set the kick angle.
        """
        return self.__angle



import numpy as np

from ..common.abstract import ReadWriteFloatArray
from ..common.abstract_aggregator import ScalarAggregator
from ..magnet.magnet import Magnet
from .element_array import ElementArray


class RWMagnetStrength(ReadWriteFloatArray):
    def __init__(self, name: str, magnets: list[Magnet]):
        self.__name = name
        self.__magnets = magnets
        self.__nb = len(self.__magnets)
        self.__aggregator: ScalarAggregator = None

    # Gets the values
    def get(self) -> np.array:
        if not self.__aggregator:
            return np.array([m.strength.get() for m in self.__magnets])
        else:
            return self.__aggregator.get()

    # Sets the values
    def set(self, value: np.array):
        nvalue = np.ones(self.__nb) * value if isinstance(value, float) else value
        if not self.__aggregator:
            for idx, m in enumerate(self.__magnets):
                m.strength.set(nvalue[idx])
        else:
            self.__aggregator.set(nvalue)

    # Sets the values and waits that the read values reach their setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the values
    def unit(self) -> list[str]:
        return [m.strength.unit() for m in self.__magnets]

    # Set the aggregator (Control system only)
    def set_aggregator(self, agg: ScalarAggregator):
        self.__aggregator = agg


class RWMagnetHardware(ReadWriteFloatArray):
    def __init__(self, name: str, magnets: list[Magnet]):
        self.__name = name
        self.__magnets = magnets
        self.__nb = len(self.__magnets)
        self.__aggregator: ScalarAggregator = None

    # Gets the values
    def get(self) -> np.array:
        if not self.__aggregator:
            return np.array([m.hardware.get() for m in self.__magnets])
        else:
            return self.__aggregator.get()

    # Sets the values
    def set(self, value: np.array):
        nvalue = np.ones(self.__nb) * value if isinstance(value, float) else value
        if not self.__aggregator:
            for idx, m in enumerate(self.__magnets):
                m.hardware.set(value[idx])
        else:
            self.__aggregator.set(value)

    # Sets the values and waits that the read values reach their setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the values
    def unit(self) -> list[str]:
        return [m.hardware.unit() for m in self.__magnets]

    # Set the aggregator
    def set_aggregator(self, agg: ScalarAggregator):
        self.__aggregator = agg


class MagnetArray(ElementArray):
    """
    Class that implements access to a magnet array

    Parameters
    ----------
    arrayName : str
        Array name
    magnets : list[Magnet]
        Magnet list, all elements must be attached to the same instance of
        either a Simulator or a ControlSystem.
    use_aggregator : bool
        Use aggregator to increase performance by using
        paralell access to underlying devices.

    Example
    -------

    An array can be retrieved from the configuration as in the following example::

        sr = Accelerator.load("acc.yaml")
        quads = sr.design.get_magnets("QuadForTune")

    """

    def __init__(self, arrayName: str, magnets: list[Magnet], use_aggregator=True):
        super().__init__(arrayName, magnets, use_aggregator)

        self.__rwstrengths = RWMagnetStrength(arrayName, magnets)
        self.__rwhardwares = RWMagnetHardware(arrayName, magnets)

        if use_aggregator and len(magnets) > 0:
            aggs = self.get_peer().create_magnet_strength_aggregator(magnets)
            aggh = self.get_peer().create_magnet_hardware_aggregator(magnets)
            self.__rwstrengths.set_aggregator(aggs)
            self.__rwhardwares.set_aggregator(aggh)

    @property
    def strengths(self) -> RWMagnetStrength:
        """
        Give access to strength of each magnet of this array
        """
        return self.__rwstrengths

    @property
    def hardwares(self) -> RWMagnetHardware:
        """
        Give access to hardware value of each magnet of this array
        """
        return self.__rwhardwares



from typing import TYPE_CHECKING

from .. import PyAMLException
from ..common.constants import ACTION_APPLY, ACTION_RESTORE
from ..common.element import Element, ElementConfigModel

if TYPE_CHECKING:
    from ..arrays.magnet_array import MagnetArray
    from ..common.element_holder import ElementHolder

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier
import json
import time
from collections.abc import Callable
from datetime import datetime

import numpy as np

# Define the main class name for this module
PYAMLCLASS = "Tune"


class ConfigModel(ElementConfigModel):
    """
    Configuration model for Tune

    Parameters
    ----------
    quad_array : str
        Array name of quad used to adjust the tune
    betatron_tune : str
        Name of the diagnostic pyaml device for measuring the tune
    delta : float
        Delta strength used to get the response matrix
    """

    quad_array: str
    betatron_tune: str
    delta: float


class TuneResponse(object):
    def __init__(self, parent: "Tune"):
        self.__parent = parent
        self.__respmatrix: np.array = None
        self.__date: str = None
        self.__correctionmat: np.array = None

    def __set_resp_map(self, mat):
        self.__respmatrix = mat
        self.__correctionmat = np.linalg.pinv(mat.T)

    def measure(self, callback: Callable | None = None):
        """
        Measure tune response matrix

        Parameters
        ----------
        callback : Callable, optional
            tune_callback(step:int,action:int,m:Magnet,dtune:np.array)
            Callback executed after each strength setting. The magnet and step are
            passed as input arguement of the callback.
            If the callback return false, then the process is aborted.
        """
        quads = self.quads()  # Returns attached quad devices
        tunemat = np.zeros((len(quads), 2))
        initial_tune = self.__parent.readback()
        delta = self.__parent._cfg.delta  # TODO: handle delta array
        aborted = False

        for idx, m in enumerate(quads):
            str = m.strength.get()
            # apply strength
            m.strength.set(str + delta)

            if callback and not callback(idx, ACTION_APPLY, m, [np.nan, np.nan]):
                aborted = True
                break

            tune = self.__parent.readback()
            dq = tune - initial_tune
            tunemat[idx] = dq / delta

            # Restore strength
            m.strength.set(str)
            if callback and not callback(idx, ACTION_RESTORE, m, tunemat[idx]):
                aborted = True
                break

        if not aborted:
            self.__set_resp_map(tunemat)
            now = datetime.now()
            self.__date = now.strftime("%m/%d/%Y, %H:%M:%S")

    def save_json(self, filename: str):
        """
        Save tune response matrix

        Parameters
        ----------
        filename: str
            File name to save
        """
        if self.__respmatrix is None:
            raise PyAMLException("TuneResponse.save(): no matrix loaded or measured")
        d = {
            "date": self.__date,
            "input_names": self.quads().names(),
            "input_delta": self.__parent._cfg.delta,
            "matrix": self.__respmatrix.tolist(),
        }
        with open(filename, "w") as f:
            json.dump(d, f)

    def load_json(self, filename: str):
        """
        Load tune response matrix

        Parameters
        ----------
        filename : str
            File name to load
        """
        with open(filename) as f:
            d = json.load(f)
        self.__date = d["date"]
        self.__input_names = d["input_names"]
        delta = d["input_delta"]
        self.__set_resp_map(np.array(d["matrix"]))

    def get(self) -> np.array:
        """
        Return tune response matrix
        """
        return self.__respmatrix

    def correct(self, dtune: np.array) -> np.array:
        """
        Return delta strengths for tune correction

        Parameters
        ----------
        dtune : np.array
            Delta tune
        """
        if self.__respmatrix is None:
            raise PyAMLException("TuneResponse.correct(): no matrix loaded or measured")
        return np.matmul(self.__correctionmat, dtune)

    def quads(self) -> "MagnetArray":
        """
        Returns quads used for tune correction
        """
        a_name = self.__parent._cfg.quad_array
        return self.__parent._peer.get_magnets(a_name)


class Tune(Element):
    """
    Class providing tune response matrix measurement and tune adjustment
    """

    def __init__(self, cfg: ConfigModel):
        """
        Construct a Tune adjustment object.

        Parameters
        ----------
        cfg : ConfigModel
            Configuration for the tune adjustment.
        """
        super().__init__(cfg.name)
        self._cfg = cfg
        self.__tm = None
        self.__tr: TuneResponse = None
        self.__setpoint = np.array([np.nan, np.nan])

    def get(self):
        """
        Return the betatron tune setpoint
        """
        return self.__setpoint

    def readback(self):
        """
        Return the betatron tune measurement
        """
        self.check_peer()
        return self.__tm.tune.get()

    def set(self, tune: np.array, iter: int = 1, wait_time: float = 0.0):
        """
        Sets the tune

        Parameters
        ----------
        tune : np.array
            Tune setpoint
        iter_nb : int
            Number of iteration
        wait_time : float
            Time to wait in second between 2 iterations
        """
        self.__setpoint = np.array(tune)
        for i in range(iter):
            diff_tune = tune - self.readback()
            str = self.__tr.quads().strengths.get()
            str += self.__tr.correct(diff_tune)
            self.__tr.quads().strengths.set(str)
            if i < iter - 1:
                # Does not wait on the last iter
                time.sleep(wait_time)

    @property
    def response(self) -> TuneResponse:
        return self.__tr

    def post_init(self):
        self.check_peer()
        self.__tm = self._peer.get_betatron_tune_monitor(self._cfg.betatron_tune)

    def attach(
        self,
        peer: "ElementHolder",
    ) -> Self:
        """
        Create a new reference to attach this tune object to a simulator
        or a control system.
        """
        obj = self.__class__(self._cfg)
        obj._peer = peer
        obj.__tr = TuneResponse(obj)
        return obj



from typing import Optional

from pydantic import BaseModel, ConfigDict

PYAMLCLASS = "ResponseMatrix"


class ConfigModel(BaseModel):
    """
    Configuration model for response matrix

    Parameters
    ----------
    matrix : list[list[float]]
        Response matrix data
    input_names : list[str], optional
        Input names
    output_names : list[str]
        Output names
    rf_response : list[float], optional
        RF response data
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    matrix: list[list[float]]
    input_names: Optional[list[str]]
    output_names: list[str]
    rf_response: Optional[list[float]]


class ResponseMatrix(object):
    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg

        # self.matrix = cfg.matrix
        # self.input_names = cfg.input_names
        # self.output_names = cfg.output_names
        # self.rf_response = cfg.rf_response



from ..common.element_holder import ElementHolder
from .array import ArrayConfig, ArrayConfigModel

# Define the main class name for this module
PYAMLCLASS = "BPM"


class ConfigModel(ArrayConfigModel): ...


class BPM(ArrayConfig):
    """
    BPM array confirguration

    Example
    -------

    A BPM array configuration can also be created by code using the following example::

        from pyaml.arrays.bpm import BPM,ConfigModel as BPMArrayConfigModel
        bpmArray = BPM(
                     BPMArrayConfigModel(name="MyBPMs", elements=["bpm1","bpm2"])
                   )
    """

    def __init__(self, cfg: ArrayConfigModel):
        super().__init__(cfg)

    def fill_array(self, holder: ElementHolder):
        holder.fill_bpm_array(self._cfg.name, self._cfg.elements)



import numpy as np
from numpy.typing import NDArray
from pydantic import BaseModel, ConfigDict

from pyaml.bpm.bpm_model import BPMModel
from pyaml.bpm.bpm_simple_model import BPMSimpleModel

from ..common.element import __pyaml_repr__
from ..control.deviceaccess import DeviceAccess

# Define the main class name for this module
PYAMLCLASS = "BPMTiltOffsetModel"


class ConfigModel(BaseModel):
    """
    Configuration model for BPM with tilt and offset

    Parameters
    ----------
    x_pos : DeviceAccess, optional
        Horizontal position device
    y_pos : DeviceAccess, optional
        Vertical position device
    x_offset : DeviceAccess, optional
        Horizontal BPM offset device
    y_offset : DeviceAccess, optional
        Vertical BPM offset device
    tilt : DeviceAccess, optional
        BPM tilt device
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    x_pos: DeviceAccess | None
    y_pos: DeviceAccess | None
    x_offset: DeviceAccess | None
    y_offset: DeviceAccess | None
    tilt: DeviceAccess | None


class BPMTiltOffsetModel(BPMSimpleModel):
    """
    Concrete implementation of BPMModel that simulates a BPM with tilt and
    offset values.
    """

    def __init__(self, cfg: ConfigModel):
        super().__init__(cfg)
        self.__x_pos = cfg.x_pos
        self.__y_pos = cfg.y_pos
        self.__x_offset = cfg.x_offset
        self.__y_offset = cfg.y_offset
        self.__tilt = cfg.tilt

    def get_pos_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles used for position reading

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        return [self.__x_pos, self.__y_pos]

    def get_tilt_device(self) -> DeviceAccess | None:
        """
        Get device handle used for tilt access

        Returns
        -------
        DeviceAccess
            DeviceAcess
        """
        return self.__tilt

    def get_offset_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles used for offset access

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        return [self.__x_offset, self.__y_offset]

    def __repr__(self):
        return __pyaml_repr__(self)



from pathlib import Path

import at
from pydantic import BaseModel, ConfigDict

from ..bpm.bpm import BPM
from ..common.abstract_aggregator import ScalarAggregator
from ..common.element import Element
from ..common.element_holder import ElementHolder
from ..common.exception import PyAMLException
from ..configuration import get_root_folder
from ..diagnostics.tune_monitor import BetatronTuneMonitor
from ..lattice.abstract_impl import (
    BPMHScalarAggregator,
    BPMScalarAggregator,
    BPMVScalarAggregator,
    RBetatronTuneArray,
    RBpmArray,
    RWBpmOffsetArray,
    RWBpmTiltScalar,
    RWHardwareArray,
    RWHardwareScalar,
    RWRFATFrequencyScalar,
    RWRFATotalVoltageScalar,
    RWRFFrequencyScalar,
    RWRFPhaseScalar,
    RWRFVoltageScalar,
    RWSerializedHardware,
    RWSerializedStrength,
    RWStrengthArray,
    RWStrengthScalar,
)
from ..magnet.cfm_magnet import CombinedFunctionMagnet
from ..magnet.magnet import Magnet
from ..magnet.serialized_magnet import SerializedMagnets
from ..rf.rf_plant import RFPlant, RWTotalVoltage
from ..rf.rf_transmitter import RFTransmitter
from ..tuning_tools.orbit import Orbit
from ..tuning_tools.tune import Tune
from .attribute_linker import (
    ConfigModel as PyAtAttrLinkerConfigModel,
)
from .attribute_linker import (
    PyAtAttributeElementsLinker,
)
from .lattice_elements_linker import LatticeElementsLinker

# Define the main class name for this module
PYAMLCLASS = "Simulator"


class ConfigModel(BaseModel):
    """
    Configuration model for Simulator

    Parameters
    ----------
    name : str
        Simulator name
    lattice : str
        AT lattice file
    mat_key : str, optional
        AT lattice ring name
    linker : LatticeElementsLinker, optional
        The linker configuration model
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    name: str
    lattice: str
    mat_key: str = None
    linker: LatticeElementsLinker = None


class Simulator(ElementHolder):
    """
    Class that implements access to AT simulator
    """

    def __init__(self, cfg: ConfigModel):
        super().__init__()
        self._cfg = cfg
        path: Path = get_root_folder() / cfg.lattice

        if self._cfg.mat_key is None:
            self.ring = at.load_lattice(path)
        else:
            self.ring = at.load_lattice(path, mat_key=f"{self._cfg.mat_key}")

        self._linker = cfg.linker
        if self._linker:
            self._linker.set_lattice(self.ring)
        else:
            self._elements_indexing = {}
            for e in self.ring:
                if e.FamName in self._elements_indexing:
                    self._elements_indexing[e.FamName].append(e)
                else:
                    self._elements_indexing[e.FamName] = [e]

    def name(self) -> str:
        return self._cfg.name

    def get_lattice(self) -> at.Lattice:
        return self.ring

    def set_energy(self, E: float):
        self.ring.energy = E
        # Needed by energy dependant element (i.e. magnet coil current calculation)
        for m in self.get_all_elements():
            m.set_energy(E)

    def create_magnet_strength_aggregator(
        self, magnets: list[Magnet]
    ) -> ScalarAggregator:
        # No magnet aggregator for simulator
        return None

    def create_magnet_hardware_aggregator(
        self, magnets: list[Magnet]
    ) -> ScalarAggregator:
        # No magnet aggregator for simulator
        return None

    def create_bpm_aggregators(self, bpms: list[BPM]) -> list[ScalarAggregator]:
        agg = BPMScalarAggregator(self.get_lattice())
        aggh = BPMHScalarAggregator(self.get_lattice())
        aggv = BPMVScalarAggregator(self.get_lattice())
        for b in bpms:
            e = self.get_at_elems(b)[0]
            agg.add_elem(e)
            aggh.add_elem(e)
            aggv.add_elem(e)
        return [agg, aggh, aggv]

    def fill_device(self, elements: list[Element]):
        for e in elements:
            # Need conversion to physics unit to work with simulator
            if isinstance(e, Magnet):
                current = (
                    RWHardwareScalar(self.get_at_elems(e), e.polynom, e.model)
                    if e.model.has_physics()
                    else None
                )
                strength = (
                    RWStrengthScalar(self.get_at_elems(e), e.polynom, e.model)
                    if e.model.has_physics()
                    else None
                )
                # Create a unique ref for this simulator
                m = e.attach(self, strength, current)
                self.add_magnet(m)

            elif isinstance(e, CombinedFunctionMagnet):
                currents = (
                    RWHardwareArray(self.get_at_elems(e), e.polynoms, e.model)
                    if e.model.has_physics()
                    else None
                )
                strengths = (
                    RWStrengthArray(self.get_at_elems(e), e.polynoms, e.model)
                    if e.model.has_physics()
                    else None
                )
                # Create unique refs of each function for this simulator
                ms = e.attach(self, strengths, currents)
                self.add_cfm_magnet(ms[0])
                for m in ms[1:]:
                    self.add_magnet(m)

            elif isinstance(e, SerializedMagnets):
                currents = []
                strengths = []
                # Create unique refs the series and each of its function for this control system
                # Link hardware to strengths and bind strength together
                for index, magnet in enumerate(e.get_magnets()):
                    current = (
                        RWHardwareScalar(
                            self.get_at_elems(magnet),
                            e.polynom,
                            e.model.get_sub_model(index),
                        )
                        if e.model.has_hardware()
                        else None
                    )
                    strength = (
                        RWStrengthScalar(
                            self.get_at_elems(magnet),
                            e.polynom,
                            e.model.get_sub_model(index),
                        )
                        if e.model.has_physics()
                        else None
                    )
                    currents.append(current)
                    strengths.append(strength)
                linked_currents = []
                linked_strengths = []
                for i in range(e.get_nb_magnets()):
                    current = (
                        RWSerializedHardware(currents, i)
                        if e.model.has_hardware()
                        else None
                    )
                    strength = (
                        RWSerializedStrength(strengths[i], currents, i)
                        if e.model.has_physics()
                        else None
                    )
                    linked_currents.append(current)
                    linked_strengths.append(strength)
                ms = e.attach(self, linked_strengths, linked_currents)
                self.add_serialized_magnet(ms[0])
                for m in ms[1:]:
                    self.add_magnet(m)

            elif isinstance(e, BPM):
                # This assumes unique BPM names in the pyAT lattice
                tilt = RWBpmTiltScalar(self.get_at_elems(e)[0])
                offsets = RWBpmOffsetArray(self.get_at_elems(e)[0])
                positions = RBpmArray(self.get_at_elems(e)[0], self.ring)
                e = e.attach(self, positions, offsets, tilt)
                self.add_bpm(e)

            elif isinstance(e, RFPlant):
                if e._cfg.transmitters:
                    cavs: list[at.Element] = []
                    harmonics: list[float] = []
                    attachedTrans: list[RFTransmitter] = []
                    for t in e._cfg.transmitters:
                        cavsPerTrans: list[at.Element] = []
                        for c in t._cfg.cavities:
                            # Expect unique name for cavities
                            cav = self.get_at_elems(Element(c))
                            if len(cav) > 1:
                                raise PyAMLException(
                                    f"RF transmitter {t.get_name()},"
                                    "multiple cavity definition:{cav[0]}"
                                )
                            if len(cav) == 0:
                                raise PyAMLException(
                                    f"RF transmitter {t.get_name()}, No cavity found"
                                )
                            cavsPerTrans.append(cav[0])
                            harmonics.append(t._cfg.harmonic)
                        voltage = RWRFVoltageScalar(cavsPerTrans)
                        phase = RWRFPhaseScalar(cavsPerTrans)
                        nt = t.attach(self, voltage, phase)
                        self.add_rf_transnmitter(nt)
                        cavs.extend(cavsPerTrans)
                        attachedTrans.append(nt)

                    frequency = RWRFFrequencyScalar(cavs, harmonics)
                    voltage = RWTotalVoltage(attachedTrans)
                    ne = e.attach(self, frequency, voltage)
                    self.add_rf_plant(ne)
                else:
                    # No transmitter defined switch to AT methods
                    frequency = RWRFATFrequencyScalar(self.ring)
                    voltage = RWRFATotalVoltageScalar(self.ring)
                    ne = e.attach(self, frequency, voltage)
                    self.add_rf_plant(ne)

            elif isinstance(e, BetatronTuneMonitor):
                betatron_tune = RBetatronTuneArray(self.ring)
                e = e.attach(self, betatron_tune)
                self.add_betatron_tune_monitor(e)

            elif isinstance(e, Tune):
                self.add_tune_tuning(e.attach(self))

            elif isinstance(e, Orbit):
                self.add_orbit_tuning(e.attach(self))

    def get_names(self, element: Element) -> list[str] | None:
        """
        Parse element lattice_name syntax. see Element.ConfigModel.lattice_name.
        """
        pattern = element.get_lattice_names()
        if pattern is None:
            return None
        if pattern.startswith("list("):
            try:
                return pattern[5:-1].rsplit(",")
            except Exception as err:
                strErr = f"{element.get_name()}: Invalid lattice_names syntax "
                strErr += f"for {pattern}, {str(err)}"
                raise PyAMLException(strErr) from err
        return None

    def get_indices(self, element: Element) -> (str | None, list[int] | None):
        """
        Parse element lattice_name syntax. see Element.ConfigModel.lattice_name.
        """

        pattern = element.get_lattice_names()
        if pattern is None:
            return (element.get_name(), None)

        # [name]@idx[,idx] syntax
        split = pattern.rfind("@")
        if split >= 0:
            try:
                name = pattern[:split]
                l = pattern[split + 1 :]
                lidx = l.rsplit(",")
                rlist = list(map(int, lidx))
                return (name if len(name) > 0 else None, rlist)
            except Exception as err:
                strErr = f"{element.get_name()}: Invalid lattice_names syntax "
                strErr += f"for {pattern}, {str(err)}"
                raise PyAMLException(strErr) from err

        # [name]#start_idx..end_idx syntax
        split = pattern.rfind("#")
        if split >= 0:
            try:
                name = pattern[:split]
                l = pattern[split + 1 :]
                lrange = l.rsplit("..")
                sl = list(map(int, lrange))
                rlist = range(sl[0], sl[1])
                return (name if len(name) > 0 else None, rlist)
            except Exception as err:
                strErr = f"{element.get_name()}: Invalid lattice_names syntax "
                strErr += f"for {pattern}, {str(err)}"
                raise PyAMLException(strErr) from err

        return (element.get_name(), None)

    def get_at_elems(self, element: Element) -> list[at.Element]:
        if self._linker:
            identifier = self._linker.get_element_identifier(element)
            element_list = self._linker.get_at_elements(identifier)
            if not element_list:
                raise PyAMLException(
                    f"{identifier} not found in lattice:{self._cfg.lattice}"
                )
            return element_list
        else:
            # By list
            nameList = self.get_names(element)

            if nameList is not None:
                names = []
                for name in nameList:
                    if name not in self._elements_indexing:
                        raise PyAMLException(
                            f"{name} not found in lattice:{self._cfg.lattice}"
                        )
                    elts = self._elements_indexing[name]
                    names.extend(elts)
                return names

            # By name or indices
            name, indices = self.get_indices(element)

            if name is None:
                # Direct indexing in the ring
                return [self.ring[idx] for idx in indices]
            else:
                if name not in self._elements_indexing:
                    raise PyAMLException(
                        f"{name} not found in lattice:{self._cfg.lattice}"
                    )
                elts = self._elements_indexing[name]
                if indices is None:
                    return elts
                else:
                    return [elts[idx] for idx in indices]

    def __repr__(self):
        return repr(self._cfg).replace("ConfigModel", self.__class__.__name__)



from ..bpm.bpm_model import BPMModel
from ..common.element import Element, ElementConfigModel
from ..common.exception import PyAMLException
from ..lattice.abstract_impl import RBpmArray, RWBpmOffsetArray, RWBpmTiltScalar

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier

PYAMLCLASS = "BPM"


class ConfigModel(ElementConfigModel):
    model: BPMModel | None = None
    """Object in charge of BPM modeling"""


class BPM(Element):
    """
    Class providing access to one BPM of a physical or simulated lattice
    """

    def __init__(self, cfg: ConfigModel):
        """
        Construct a BPM

        Parameters
        ----------
        name : str
            Element name
        hardware : DeviceAccess
            Direct access to a hardware (bypass the BPM model)
        model : BPMModel
            BPM model in charge of computing beam position
        """

        super().__init__(cfg.name)

        self.__model = cfg.model if hasattr(cfg, "model") else None
        self._cfg = cfg
        self.__positions = None
        self.__offset = None
        self.__tilt = None

    @property
    def model(self) -> BPMModel:
        return self.__model

    @property
    def positions(self) -> RBpmArray:
        if self.__positions is None:
            raise PyAMLException(f"{str(self)} has no attached positions")
        return self.__positions

    @property
    def offset(self) -> RWBpmOffsetArray:
        if self.__offset is None:
            raise PyAMLException(f"{str(self)} has no attached offset")
        return self.__offset

    @property
    def tilt(self) -> RWBpmTiltScalar:
        if self.__tilt is None:
            raise PyAMLException(f"{str(self)} has no attached tilt")
        return self.__tilt

    def attach(
        self,
        peer,
        positions: RBpmArray,
        offset: RWBpmOffsetArray,
        tilt: RWBpmTiltScalar,
    ) -> Self:
        # Attach positions, offset and tilt attributes and returns a new
        # reference
        obj = self.__class__(self._cfg)
        obj.__model = self.__model
        obj.__positions = positions
        obj.__offset = offset
        obj.__tilt = tilt
        obj._peer = peer
        return obj



from abc import ABCMeta, abstractmethod

import numpy as np
import numpy.typing as npt

from ..control.deviceaccess import DeviceAccess


class MagnetModel(metaclass=ABCMeta):
    """
    Abstract class providing strength to coil current conversion
    and access to underlying power supplies
    """

    @abstractmethod
    def compute_hardware_values(
        self, strengths: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """
        Compute hardware value(s) from magnet strength(s)

        Parameters
        ----------
        strengths : npt.NDArray[np.float64]
            Array of strengths. For a single multipole,
            strengths is an array of 1 item.

        Returns
        -------
        npt.NDArray[np.float64]
            Array of hardware values (i.e. currents or voltages).
        """
        pass

    @abstractmethod
    def compute_strengths(
        self, hardware_values: npt.NDArray[np.float64]
    ) -> npt.NDArray[np.float64]:
        """
        Compute magnet strength(s) from hardware value(s)

        Parameters
        ----------
        hardware_values : npt.NDArray[np.float64]
            Array of hardware values (i.e. currents or voltages)

        Returns
        -------
        npt.NDArray[np.float64]
            Array of strengths. For a single multipole,
            returns an array of 1 item
        """
        pass

    @abstractmethod
    def get_strength_units(self) -> list[str]:
        """
        Get strength units

        Returns
        -------
        list[str]
            Array of strength units. For a single multipole,
            returns a list of 1 item
        """
        pass

    @abstractmethod
    def get_hardware_units(self) -> list[str]:
        """
        Get hardware units

        Returns
        -------
        list[str]
            Array of hardware units. For a single multipole,
            returns a list of 1 item
        """
        pass

    @abstractmethod
    def get_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        pass

    @abstractmethod
    def set_magnet_rigidity(self, brho: np.double):
        """
        Set magnet rigidity

        Parameters
        ----------
        brho: np.double
            Magnet rigidity used to calculate power supply setpoints
        """
        pass

    def has_hardware(self) -> bool:
        """
        Tells if the model allows to work in hardware unit.

        Returns
        ----------
        bool
            True if the model supports hardware unit
        """
        return True

    def has_physics(self) -> bool:
        """
        Tells if the model allows to work in physics unit.

        Returns
        ----------
        bool
            True if the model supports physics unit
        """
        return True



from scipy.constants import speed_of_light

from .. import PyAMLException
from ..common import abstract
from ..common.element import Element, ElementConfigModel
from .model import MagnetModel

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier
import numpy as np


class MagnetConfigModel(ElementConfigModel):
    model: MagnetModel | None = None
    """Object in charge of converting magnet strenghts to power supply values"""


class Magnet(Element):
    """
    Class providing access to one magnet of a physical or simulated lattice
    """

    def __init__(self, name: str, model: MagnetModel = None):
        """
        Construct a magnet

        Parameters
        ----------
        name : str
            Element name
        model : MagnetModel
            Magnet model in charge of computing coil(s) current
        """
        super().__init__(name)
        self.__model = model
        self.__strength: abstract.ReadWriteFloatScalar = None
        self.__hardware: abstract.ReadWriteFloatScalar = None
        self.__modelName = self.get_name()

    @property
    def strength(self) -> abstract.ReadWriteFloatScalar:
        """
        Gives access to the strength of this magnet in physics unit
        """
        self.check_peer()
        if self.__strength is None:
            raise PyAMLException(
                f"{str(self)} has no model that supports physics units"
            )
        return self.__strength

    @property
    def hardware(self) -> abstract.ReadWriteFloatScalar:
        """
        Gives access to the strength of this magnet in
        hardware unit when possible
        """
        self.check_peer()
        if self.__hardware is None:
            raise PyAMLException(
                f"{str(self)} has no model that supports hardware units"
            )
        return self.__hardware

    @property
    def model(self) -> MagnetModel:
        """
        Returns a handle to the underlying magnet model
        """
        return self.__model

    def attach(
        self,
        peer,
        strength: abstract.ReadWriteFloatScalar,
        hardware: abstract.ReadWriteFloatScalar,
    ) -> Self:
        """
        Create a new reference to attach this magnet to a simulator
        or a control systemand.
        """
        obj = self.__class__(self._cfg)
        obj.__modelName = self.__modelName
        obj.__strength = strength
        obj.__hardware = hardware
        obj._peer = peer
        return obj

    def set_energy(self, energy: float):
        """
        Set the energy in eV to compute and set the magnet rigidity
        on the underlying magnet model.
        """
        if self.__model is not None:
            self.__model.set_magnet_rigidity(np.double(energy / speed_of_light))

    def set_model_name(self, name: str):
        """
        Sets the name of this magnet in the model
        (Used for combined function manget)
        """
        self.__modelName = name

    def get_model_name(self) -> str:
        """
        Returns the model name of this magnet
        """
        return self.__modelName

    def __repr__(self):
        return "%s(peer='%s', name='%s', model_name='%s', magnet_model=%s)" % (
            self.__class__.__name__,
            self.get_peer(),
            self.get_name(),
            self.__modelName,
            repr(self.__model),
        )



"""
pyAML package
~~~~~~~~~~~~~

pyAML
"""

__title__ = "pyAML"
__description__ = "Python Accelerator Middle Layer"
__url__ = "https://github.com/python-accelerator-middle-layer/pyaml"
__version__ = "0.2.1"
__author__ = "pyAML collaboration"
__author_email__ = ""

import logging.config
import os

from pyaml.common.exception import PyAMLConfigException, PyAMLException

__all__ = [__version__, "PyAMLException", "PyAMLConfigException"]


config_file = os.getenv("PYAML_LOG_CONFIG", "pyaml_logging.conf")

if os.path.exists(config_file):
    logging.config.fileConfig(config_file, disable_existing_loggers=False)

logger = logging.getLogger("pyaml")
level = os.getenv("PYAML_LOG_LEVEL", "").upper()
if len(level) > 0:
    logger.setLevel(getattr(logging, level, logging.WARNING))



import logging
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, ConfigDict

from ..common.element_holder import ElementHolder
from ..common.exception import PyAMLException
from ..external.pySC import pySC
from ..external.pySC.pySC.apps import measure_ORM
from ..external.pySC.pySC.apps.codes import ResponseCode
from ..external.pySC_interface import pySCInterface

logger = logging.getLogger(__name__)

PYAMLCLASS = "OrbitResponseMatrix"


class ConfigModel(BaseModel):
    """
    Configuration model for orbit response matrix measurement

    Parameters
    ----------
    bpm_array_name : str
        BPM array name
    hcorr_array_name : str
        Horizontal corrector array name
    vcorr_array_name : str
        Vertical corrector array name
    corrector_delta : float
        Corrector delta for measurement
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    bpm_array_name: str
    hcorr_array_name: str
    vcorr_array_name: str
    corrector_delta: float


class OrbitResponseMatrix(object):
    def __init__(self, element_holder: ElementHolder, cfg: ConfigModel):
        self._cfg = cfg

        self.element_holder = element_holder
        self.bpm_array_name = cfg.bpm_array_name
        self.hcorr_array_name = cfg.hcorr_array_name
        self.vcorr_array_name = cfg.vcorr_array_name
        self.corrector_delta = cfg.corrector_delta
        self.latest_measurement = None

    def measure(self, corrector_names: Optional[List[str]] = None):
        interface = pySCInterface(
            element_holder=self.element_holder,
            bpm_array_name=self.bpm_array_name,
        )

        if corrector_names is None:
            logger.info(
                f"Measuring correctors from the default arrays: "
                f"{self.hcorr_array_name} and {self.vcorr_array_name}."
            )
            hcorrector_names = self.element_holder.get_magnets(
                self.hcorr_array_name
            ).names()
            vcorrector_names = self.element_holder.get_magnets(
                self.vcorr_array_name
            ).names()
            corrector_names = hcorrector_names + vcorrector_names

        generator = measure_ORM(
            interface=interface,
            corrector_names=corrector_names,
            delta=self.corrector_delta,
            skip_save=True,
        )

        pySC.disable_pySC_rich()
        for code, measurement in generator:
            if code is ResponseCode.MEASURING:
                logger.info(f"Measured response of {measurement.last_input}.")

        response_data = measurement.response_data  # contains also pre-processed data
        response_data.output_names = self.element_holder.get_bpms(
            self.bpm_array_name
        ).names()
        self.latest_measurement = response_data.model_dump()

    def get(self):
        return self.latest_measurement

    def save(self, save_path: Path, with_type: str = "json"):
        # should we make a general pyaml saving/loading function for data?
        if with_type == "json":
            import json

            data = self.latest_measurement
            json.dump(data, open(save_path, "w"), indent=4)
        elif with_type == "yaml":
            import yaml

            data = self.latest_measurement
            yaml.safe_dump(data, open(save_path, "w"))
        elif with_type == "npz":
            import numpy as np

            data = self.latest_measurement
            np.savez(save_path.resolve(), **data)
        else:
            raise PyAMLException(f"ERROR: Unknown file type to save as: {with_type}.")



from abc import ABCMeta, abstractmethod

import numpy as np
import numpy.typing as npt


class ScalarAggregator(metaclass=ABCMeta):
    """
    Abstract class providing access to a list of scalar variables
    """

    @abstractmethod
    def set(self, value: npt.NDArray[np.float64]):
        """Write a list of variable"""
        pass

    @abstractmethod
    def set_and_wait(self, value: npt.NDArray[np.float64]):
        """Write a list of variable and wait that setpoint are reached"""
        pass

    @abstractmethod
    def get(self) -> npt.NDArray[np.float64]:
        """Return a list variables"""
        pass

    @abstractmethod
    def readback(self) -> np.array:
        """Return  a list variables (measurements)"""
        pass

    @abstractmethod
    def unit(self) -> str:
        """Return the variables unit"""
        pass



from ..lattice.polynom_info import PolynomInfo
from .magnet import Magnet, MagnetConfigModel

# Define the main class name for this module
PYAMLCLASS = "SkewOctu"


class ConfigModel(MagnetConfigModel): ...


class SkewOctu(Magnet):
    """SkewOctu class"""

    polynom = PolynomInfo("PolynomA", 3)

    def __init__(self, cfg: ConfigModel):
        super().__init__(
            cfg.name,
            cfg.model if hasattr(cfg, "model") else None,
        )
        self._cfg = cfg



from abc import ABCMeta, abstractmethod

import numpy as np
import numpy.typing as npt

from .deviceaccess import DeviceAccess
from .readback_value import Value


class DeviceAccessList(list[DeviceAccess], metaclass=ABCMeta):
    """
    Abstract class providing access to a list of control system float variable
    """

    @abstractmethod
    def add_devices(self, devices: DeviceAccess | list[DeviceAccess]):
        """Add a DeviceAccess to this list"""
        pass

    @abstractmethod
    def get_devices(self) -> DeviceAccess | list[DeviceAccess]:
        """Get the DeviceAccess list"""
        pass

    @abstractmethod
    def set(self, value: npt.NDArray[np.float64]):
        """Write a list of control system device variable
        (i.e. a power supply currents)"""
        pass

    @abstractmethod
    def set_and_wait(self, value: npt.NDArray[np.float64]):
        """Write a list control system device variable (i.e. a power supply currents)"""
        pass

    @abstractmethod
    def get(self) -> npt.NDArray[np.float64]:
        """Return a list of setpoints of control system device variables"""
        pass

    @abstractmethod
    def readback(self) -> np.array:
        """Return the measured variables"""
        pass

    @abstractmethod
    def unit(self) -> str:
        """Return the variable unit"""
        pass



import json
from pathlib import Path

import numpy as np

from pyaml.accelerator import Accelerator
from pyaml.tuning_tools.dispersion import ConfigModel as Disp_ConfigModel
from pyaml.tuning_tools.dispersion import Dispersion
from pyaml.tuning_tools.orbit_response_matrix import ConfigModel as ORM_ConfigModel
from pyaml.tuning_tools.orbit_response_matrix import OrbitResponseMatrix

parent_folder = Path(__file__).parent
config_path = parent_folder.parent.parent.joinpath(
    "tests", "config", "EBSOrbit.yaml"
).resolve()
sr = Accelerator.load(config_path)
element_holder = sr.design

orm = OrbitResponseMatrix(
    cfg=ORM_ConfigModel(
        bpm_array_name="BPM",
        hcorr_array_name="HCorr",
        vcorr_array_name="VCorr",
        corrector_delta=1e-6,
    ),
    element_holder=element_holder,
)

orm.measure()

orm_data = orm.get()

dispersion = Dispersion(
    cfg=Disp_ConfigModel(
        bpm_array_name="BPM",
        rf_plant_name="RF",
        frequency_delta=10,
    ),
    element_holder=element_holder,
)

dispersion.measure()
dispersion_data = dispersion.get()
rf_response = (
    dispersion_data["frequency_response_x"] + dispersion_data["frequency_response_y"]
)

ideal_ORM_data = {
    "type": "pyaml.tuning_tools.response_matrix",
    "matrix": orm_data["matrix"],
    "input_names": orm_data["input_names"],
    "output_names": orm_data["output_names"],
    "rf_response": rf_response,
}

json.dump(ideal_ORM_data, open("ideal_orm_disp.json", "w"))



import at
import numpy as np
from numpy.typing import NDArray
from scipy.constants import speed_of_light

from ..common import abstract
from ..common.abstract_aggregator import ScalarAggregator
from ..common.exception import PyAMLException
from ..magnet.model import MagnetModel
from ..rf.rf_plant import RFPlant
from ..rf.rf_transmitter import RFTransmitter
from .polynom_info import PolynomInfo

# TODO handle serialized magnets for magnet array

# ------------------------------------------------------------------------------


class RWHardwareScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a magnet of a simulator in hardware unit.
    Hardware unit is converted from strength using the magnet model
    """

    def __init__(
        self, elements: list[at.Element], poly: PolynomInfo, model: MagnetModel
    ):
        self.__model = model
        self.__elements = elements
        self.__poly = [e.__getattribute__(poly.attName) for e in elements]
        self.__sign = poly.sign
        self.__polyIdx = poly.index
        self.__length = 0
        for e in elements:
            self.__length += e.Length

    def get(self) -> float:
        s = 0
        for idx, e in enumerate(self.__elements):
            s += self.__poly[idx][self.__polyIdx] * self.__sign * e.Length
        return self.__model.compute_hardware_values([s])[0]

    def set(self, value: float):
        s = self.__model.compute_strengths([value])[0]
        for idx, _ in enumerate(self.__elements):
            self.__poly[idx][self.__polyIdx] = s / (self.__length * self.__sign)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self.__model.get_hardware_units()[0]

    def get_model(self) -> MagnetModel:
        return self.__model


# ------------------------------------------------------------------------------


class RWStrengthScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a strength of a simulator
    """

    def __init__(
        self, elements: list[at.Element], poly: PolynomInfo, model: MagnetModel
    ):
        self.__model = model
        self.__elements = elements
        self.__poly = [e.__getattribute__(poly.attName) for e in elements]
        self.__sign = poly.sign
        self.__polyIdx = poly.index
        self.__length = 0
        for e in elements:
            self.__length += e.Length

    # Gets the value
    def get(self) -> float:
        s = 0
        for idx, e in enumerate(self.__elements):
            s += self.__poly[idx][self.__polyIdx] * self.__sign * e.Length
        return s

    # Sets the value
    def set(self, value: float):
        for idx, _ in enumerate(self.__elements):
            self.__poly[idx][self.__polyIdx] = value / (self.__length * self.__sign)

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return self.__model.get_strength_units()[0]

    # ------------------------------------------------------------------------------
    def get_model(self) -> MagnetModel:
        return self.__model


# ------------------------------------------------------------------------------


class RWSerializedHardware(abstract.ReadWriteFloatScalar):
    def __init__(self, elements: list[RWHardwareScalar], element_index: int):
        self.__elements = elements
        self.__element_index = element_index

    # Gets the value
    def get(self) -> float:
        return self.__elements[self.__element_index].get()

    # Sets the value
    def set(self, value: float):
        [element.set(value) for element in self.__elements]

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return self.__elements[self.__element_index].unit()

    def set_magnet_rigidity(self, brho: np.double):
        [element.get_model().set_magnet_rigidity(brho) for element in self.__elements]


class RWSerializedStrength(abstract.ReadWriteFloatScalar):
    def __init__(
        self,
        element: RWStrengthScalar,
        elements_hardware: list[RWHardwareScalar],
        element_index: int,
    ):
        self.__element = element
        self.__elements_hardware = elements_hardware
        self.__element_index = element_index

    # Gets the value
    def get(self) -> float:
        return self.__element.get()

    # Sets the value
    def set(self, value: float):
        self.__element.set(value)
        hardware_value = self.__elements_hardware[self.__element_index].get()
        [
            element.set(hardware_value)
            for index, element in enumerate(self.__elements_hardware)
            if index != self.__element_index
        ]

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return self.__element.unit()

    def set_magnet_rigidity(self, brho: np.double):
        pass


class RWHardwareArray(abstract.ReadWriteFloatArray):
    """
    Class providing read write access to a magnet of a simulator in hardware units.
    Hardware units are converted from strengths using the magnet model
    """

    def __init__(
        self, elements: list[at.Element], poly: list[PolynomInfo], model: MagnetModel
    ):
        self.__elements = elements
        self.__poly = []
        self.__polyIdx = []
        self.__sign = []
        self.__model = model
        for p in poly:
            self.__poly.append(elements[0].__getattribute__(p.attName))
            self.__polyIdx.append(p.index)
            self.__sign.append(p.sign)

    # Gets the value
    def get(self) -> np.array:
        nbStrength = len(self.__poly)
        s = np.zeros(nbStrength)
        for i in range(nbStrength):
            s[i] = (
                self.__poly[i][self.__polyIdx[i]]
                * self.__sign[i]
                * self.__elements[0].Length
            )
        return self.__model.compute_hardware_values(s)

    # Sets the value
    def set(self, value: np.array):
        nbStrength = len(self.__poly)
        s = self.__model.compute_strengths(value)
        for i in range(nbStrength):
            self.__poly[i][self.__polyIdx[i]] = s[i] / (
                self.__elements[0].Length * self.__sign[i]
            )

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> list[str]:
        return self.__model.get_hardware_units()


# ------------------------------------------------------------------------------


class RWStrengthArray(abstract.ReadWriteFloatArray):
    """
    Class providing read write access to a strength (array) of a simulator
    """

    def __init__(
        self, elements: list[at.Element], poly: list[PolynomInfo], model: MagnetModel
    ):
        self.__elements = elements
        self.__poly = []
        self.__polyIdx = []
        self.__sign = []
        self.__model = model
        for p in poly:
            self.__poly.append(elements[0].__getattribute__(p.attName))
            self.__polyIdx.append(p.index)
            self.__sign.append(p.sign)

    # Gets the value
    def get(self) -> np.array:
        nbStrength = len(self.__poly)
        s = np.zeros(nbStrength)
        for i in range(nbStrength):
            s[i] = (
                self.__poly[i][self.__polyIdx[i]]
                * self.__sign[i]
                * self.__elements[0].Length
            )
        return s

    # Sets the value
    def set(self, value: np.array):
        nbStrength = len(self.__poly)
        s = np.zeros(nbStrength)
        for i in range(nbStrength):
            self.__poly[i][self.__polyIdx[i]] = value[i] / (
                self.__elements[0].Length * self.__sign[i]
            )

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> list[str]:
        return self.__model.get_strength_units()


# ------------------------------------------------------------------------------


class BPMScalarAggregator(ScalarAggregator):
    """
    BPM simulator aggregator
    """

    def __init__(self, ring: at.Lattice):
        self.lattice = ring
        self.refpts = []

    def add_elem(self, elem: at.Element):
        self.refpts.append(self.lattice.index(elem))

    def set(self, value: NDArray[np.float64]):
        pass

    def set_and_wait(self, value: NDArray[np.float64]):
        pass

    def get(self) -> np.array:
        _, orbit = at.find_orbit(self.lattice, refpts=self.refpts)
        return orbit[:, [0, 2]].flatten()

    def readback(self) -> np.array:
        return self.get()

    def unit(self) -> str:
        return "m"


# ------------------------------------------------------------------------------


class BPMHScalarAggregator(BPMScalarAggregator):
    """
    Vertical BPM simulator aggregator
    """

    def get(self) -> np.array:
        _, orbit = at.find_orbit(self.lattice, refpts=self.refpts)
        return orbit[:, 0]


# ------------------------------------------------------------------------------


class BPMVScalarAggregator(BPMScalarAggregator):
    """
    Horizontal BPM simulator aggregator
    """

    def get(self) -> np.array:
        _, orbit = at.find_orbit(self.lattice, refpts=self.refpts)
        return orbit[:, 2]


# ------------------------------------------------------------------------------


class RBpmArray(abstract.ReadFloatArray):
    """
    Class providing read access to a BPM position (array) of a simulator.
    Position in pyAT is calculated using find_orbit function, which returns the
    orbit at a specified index. The position is then extracted from the orbit
    array as the first two elements (x, y).
    """

    def __init__(self, element: at.Element, lattice: at.Lattice):
        self.__element = element
        self.__lattice = lattice

    # Gets the value
    def get(self) -> np.array:
        index = self.__lattice.index(self.__element)
        _, orbit = at.find_orbit(self.__lattice, refpts=index)
        return orbit[0, [0, 2]]

    # Gets the unit of the value
    def unit(self) -> str:
        return "m"


# ------------------------------------------------------------------------------


class RWBpmOffsetArray(abstract.ReadWriteFloatArray):
    """
    Class providing read write access to a BPM offset (array) of a simulator.
    Offset in pyAT is defined in Offset attribute as a 2-element array.
    """

    def __init__(self, element: at.Element):
        self.__element = element
        try:
            self.__offset = element.__getattribute__("Offset")
        except AttributeError:
            self.__offset = None

    # Gets the value
    def get(self) -> np.array:
        if self.__offset is None:
            raise PyAMLException("Element does not have an Offset attribute.")
        return self.__offset

    # Sets the value
    def set(self, value: np.array):
        if self.__offset is None:
            raise PyAMLException("Element does not have an Offset attribute.")
        if len(value) != 2:
            raise PyAMLException("BPM offset must be a 2-element array.")
        self.__offset = value

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return "m"  # Assuming all offsets are in m


# ------------------------------------------------------------------------------


class RWBpmTiltScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a BPM tilt of a simulator. Tilt in
    pyAT is defined in Rotation attribute as a first element.
    """

    def __init__(self, element: at.Element):
        self.__element = element
        try:
            self.__tilt = element.__getattribute__("Rotation")[0]
        except AttributeError:
            self.__tilt = None

    # Gets the value
    def get(self) -> float:
        if self.__tilt is None:
            raise ValueError("Element does not have a Tilt attribute.")
        return self.__tilt

    # Sets the value
    def set(
        self,
        value: float,
    ):
        self.__tilt = value
        self.__element.__setattr__("Rotation", [value, None, None])

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return "rad"  # Assuming BPM tilts are in rad


# ------------------------------------------------------------------------------


class RWRFVoltageScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a cavity voltage
    of a simulator for a given RF trasnmitter.
    """

    def __init__(self, elements: list[at.Element]):
        self.__elements = elements

    def get(self) -> float:
        sum = 0
        for _idx, e in enumerate(self.__elements):
            sum += e.Voltage
        return sum

    def set(self, value: float):
        v = value / len(self.__elements)
        for e in self.__elements:
            e.Voltage = v

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return "V"


# ------------------------------------------------------------------------------


class RWRFPhaseScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a cavity phase of
    a simulator for a given RF trasnmitter.
    """

    def __init__(self, elements: list[at.Element]):
        self.__elements = elements

    def get(self) -> float:
        # Assume that all cavities of this transmitter
        # have the same Time Lag and Frequency
        wavelength = speed_of_light / self.__elements[0].Frequency
        return (wavelength / self.__elements[0].TimeLag) * 2.0 * np.pi

    def set(self, value: float):
        wavelength = speed_of_light / self.__elements[0].Frequency
        for e in self.__elements:
            e.TimeLag = wavelength * value / (2.0 * np.pi)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return "rad"


# ------------------------------------------------------------------------------


class RWRFFrequencyScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to RF frequency of a simulator.
    """

    def __init__(self, elements: list[at.Element], harmonics: list[float]):
        self.__elements = elements
        self.__harm = harmonics

    def get(self) -> float:
        # Serialized cavity has the same frequency
        return self.__elements[0].Frequency

    def set(self, value: float):
        for idx, e in enumerate(self.__elements):
            e.Frequency = value * self.__harm[idx]

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return "Hz"


# ------------------------------------------------------------------------------


class RWRFATFrequencyScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to RF frequency of a simulator using
    AT methods.
    """

    def __init__(self, ring: at.Lattice):
        self.__ring = ring

    def get(self) -> float:
        return self.__ring.get_rf_frequency()

    def set(self, value: float):
        self.__ring.set_rf_frequency(value)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return "Hz"


# ------------------------------------------------------------------------------


class RWRFATotalVoltageScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a RF voltage of a simulator using AT methods.
    """

    def __init__(self, ring: at.Lattice):
        self.__ring = ring

    def get(self) -> float:
        return self.__ring.get_rf_voltage()

    def set(self, value: float):
        self.__ring.set_rf_voltage(value)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return "V"


# ------------------------------------------------------------------------------


class RBetatronTuneArray(abstract.ReadFloatArray):
    """
    Class providing read-only access to the betatron tune of a ring.
    """

    def __init__(self, ring: at.Lattice):
        self.__ring = ring

    def get(self) -> float:
        return self.__ring.get_tune()[:2]

    def unit(self) -> str:
        return "1"



"""
Module handling element references for simulators and control system
"""

from typing import TYPE_CHECKING

from ..arrays.bpm_array import BPMArray
from ..arrays.cfm_magnet_array import CombinedFunctionMagnetArray
from ..arrays.element_array import ElementArray
from ..arrays.magnet_array import MagnetArray
from ..arrays.serialized_magnet_array import SerializedMagnetsArray
from ..bpm.bpm import BPM
from ..common.exception import PyAMLException
from ..diagnostics.tune_monitor import BetatronTuneMonitor
from ..magnet.cfm_magnet import CombinedFunctionMagnet
from ..magnet.magnet import Magnet
from ..magnet.serialized_magnet import SerializedMagnets
from ..rf.rf_plant import RFPlant
from ..rf.rf_transmitter import RFTransmitter
from .element import Element

if TYPE_CHECKING:
    from ..tuning_tools.orbit import Orbit
    from ..tuning_tools.tune import Tune


class ElementHolder(object):
    """
    Class that store references of objects used from both
    simulators and control system
    """

    def __init__(self):
        # Device handle
        self.__MAGNETS: dict = {}
        self.__CFM_MAGNETS: dict = {}
        self.__SERIALIZED_MAGNETS: dict = {}
        self.__BPMS: dict = {}
        self.__RFPLANT: dict = {}
        self.__RFTRANSMITTER: dict = {}
        self.__DIAG: dict = {}
        self.__TUNING_TOOLS = {}
        self.__ALL: dict = {}

        # Array handle
        self.__MAGNET_ARRAYS: dict = {}
        self.__CFM_MAGNET_ARRAYS: dict = {}
        self.__SERIALIZED_MAGNETS_ARRAYS: dict = {}
        self.__BPM_ARRAYS: dict = {}
        self.__ELEMENT_ARRAYS: dict = {}

    def post_init(self):
        """
        Method triggered after all initialisations are done
        """
        for e in self.get_all_elements():
            e.post_init()

    def fill_device(self, elements: list[Element]):
        raise PyAMLException("ElementHolder.fill_device() is not subclassed")

    def fill_array(
        self, arrayName: str, elementNames: list[str], get_func, constructor, ARR: dict
    ):
        a = []
        for name in elementNames:
            try:
                m = get_func(name)
            except Exception as err:
                raise PyAMLException(
                    f"{constructor.__name__} {arrayName} : {err} @index {len(a)}"
                ) from None
            if m in a:
                raise PyAMLException(
                    f"{constructor.__name__} {arrayName} : "
                    f"duplicate name {name} @index {len(a)}"
                ) from None
            a.append(m)
        ARR[arrayName] = constructor(arrayName, a)

    def __add(self, array, element: Element):
        if element.get_name() in self.__ALL:  # Ensure name unicity
            raise PyAMLException(
                f"Duplicate element {element.__class__.__name__} "
                "name {element.get_name()}"
            ) from None
        array[element.get_name()] = element
        self.__ALL[element.get_name()] = element

    def __get(self, what, name, array) -> Element:
        if name not in array:
            raise PyAMLException(f"{what} {name} not defined")
        return array[name]

    # Generic elements
    def fill_element_array(self, arrayName: str, elementNames: list[str]):
        self.fill_array(
            arrayName,
            elementNames,
            self.get_element,
            ElementArray,
            self.__ELEMENT_ARRAYS,
        )

    def get_element(self, name: str) -> Element:
        return self.__get("Element", name, self.__ALL)

    def get_elements(self, name: str) -> ElementArray:
        return self.__get("Element array", name, self.__ELEMENT_ARRAYS)

    def get_all_elements(self) -> list[Element]:
        return [value for key, value in self.__ALL.items()]

    # Magnets

    def fill_magnet_array(self, arrayName: str, elementNames: list[str]):
        self.fill_array(
            arrayName, elementNames, self.get_magnet, MagnetArray, self.__MAGNET_ARRAYS
        )

    def get_magnet(self, name: str) -> Magnet:
        return self.__get("Magnet", name, self.__MAGNETS)

    def add_magnet(self, m: Magnet):
        self.__add(self.__MAGNETS, m)

    def get_magnets(self, name: str) -> MagnetArray:
        return self.__get("Magnet array", name, self.__MAGNET_ARRAYS)

    def get_all_magnets(self) -> list[Magnet]:
        return [value for key, value in self.__MAGNETS.items()]

    # Combined Function Magnets

    def fill_cfm_magnet_array(self, arrayName: str, elementNames: list[str]):
        self.fill_array(
            arrayName,
            elementNames,
            self.get_cfm_magnet,
            CombinedFunctionMagnetArray,
            self.__CFM_MAGNET_ARRAYS,
        )

    def get_cfm_magnet(self, name: str) -> Magnet:
        return self.__get("CombinedFunctionMagnet", name, self.__CFM_MAGNETS)

    def add_cfm_magnet(self, m: Magnet):
        self.__add(self.__CFM_MAGNETS, m)

    def get_cfm_magnets(self, name: str) -> CombinedFunctionMagnetArray:
        return self.__get(
            "CombinedFunctionMagnet array", name, self.__CFM_MAGNET_ARRAYS
        )

    def get_all_cfm_magnets(self) -> list[CombinedFunctionMagnet]:
        return [value for key, value in self.__CFM_MAGNETS.items()]

    # Serialized magnets

    def fill_serialized_magnet_array(self, arrayName: str, elementNames: list[str]):
        self.fill_array(
            arrayName,
            elementNames,
            self.get_serialized_magnet,
            SerializedMagnetsArray,
            self.__SERIALIZED_MAGNETS_ARRAYS,
        )

    def get_serialized_magnet(self, name: str) -> Magnet:
        return self.__get("SerializedMagnets", name, self.__SERIALIZED_MAGNETS)

    def add_serialized_magnet(self, m: Magnet):
        self.__add(self.__SERIALIZED_MAGNETS, m)

    def get_serialized_magnets(self, name: str) -> SerializedMagnetsArray:
        return self.__get(
            "SerializedMagnets array", name, self.__SERIALIZED_MAGNETS_ARRAYS
        )

    def get_all_serialized_magnets(self) -> list[SerializedMagnets]:
        return [value for key, value in self.__SERIALIZED_MAGNETS.items()]

    # BPMs

    def fill_bpm_array(self, arrayName: str, elementNames: list[str]):
        self.fill_array(
            arrayName, elementNames, self.get_bpm, BPMArray, self.__BPM_ARRAYS
        )

    def get_bpm(self, name: str) -> Element:
        return self.__get("BPM", name, self.__BPMS)

    def add_bpm(self, bpm: BPM):
        self.__add(self.__BPMS, bpm)

    def get_bpms(self, name: str) -> BPMArray:
        return self.__get("BPM array", name, self.__BPM_ARRAYS)

    def get_all_bpms(self) -> list[BPM]:
        return [value for key, value in self.__BPMS.items()]

    # RF

    def get_rf_plant(self, name: str) -> RFPlant:
        return self.__get("RFPlant", name, self.__RFPLANT)

    def add_rf_plant(self, rf: RFPlant):
        self.__add(self.__RFPLANT, rf)

    def add_rf_transnmitter(self, rf: RFTransmitter):
        self.__add(self.__RFTRANSMITTER, rf)

    def get_rf_trasnmitter(self, name: str) -> RFTransmitter:
        return self.__get("RFTransmitter", name, self.__RFTRANSMITTER)

    # Tune monitor

    def get_betatron_tune_monitor(self, name: str) -> BetatronTuneMonitor:
        return self.__get("Diagnostic", name, self.__DIAG)

    def add_betatron_tune_monitor(self, tune_monitor: Element):
        self.__add(self.__DIAG, tune_monitor)

    # Tuning tools

    def get_tune_tuning(self, name: str) -> "Tune":
        return self.__get("Tune tuning tool", name, self.__TUNING_TOOLS)

    def add_tune_tuning(self, tune: Element):
        self.__add(self.__TUNING_TOOLS, tune)

    @property
    def tune(self) -> "Tune":
        return self.get_tune_tuning("DEFAULT_TUNE_CORRECTION")

    def get_orbit_tuning(self, name: str) -> "Orbit":
        return self.__get("Orbit tuning tool", name, self.__TUNING_TOOLS)

    def add_orbit_tuning(self, orbit: Element):
        self.__add(self.__TUNING_TOOLS, orbit)

    @property
    def orbit(self) -> "Orbit":
        return self.get_orbit_tuning("DEFAULT_ORBIT_CORRECTION")



"""
slow orbit feedback package
~~~~~~~~~~~~~~~~

pyAML

"""

__title__ = "pyAML"
__description__ = "Python Accelerator Middle Layer"
__url__ = "https://github.com/python-accelerator-middle-layer/pyaml"
__version__ = "0.0.0"
__author__ = "pyAML collaboration"
__author_email__ = ""

__all__ = [__version__]



from abc import ABCMeta, abstractmethod

from numpy import array


class Matrix(metaclass=ABCMeta):
    """
    Abstract class providing access to a matrix
    """

    @abstractmethod
    def get_matrix(self) -> array:
        """Returns the matrix"""
        pass



class PyAMLException(Exception):
    """Exception raised for PyAML error scenarios.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message


class PyAMLConfigException(Exception):
    """Exception raised for PyAML configuration error scenarios.

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message):
        super().__init__(message)
        self.message = message



import numpy as np

from ..common import abstract
from .magnet import Magnet


class RWCorrectorAngle(abstract.ReadWriteFloatScalar):
    """
    Set the angle of a horizontal or vertical corrector.
    KickAngle sign convention is defined the a global PyAML constant
    (see pyaml.common.constant.HORIZONATL_KICK_SIGN).
    To change the convention, you have execute the code below prior to everything:
    import pyaml.common.constants
    pyaml.common.constants.HORIZONTAL_KICK_SIGN = -1.0
    """

    def __init__(self, corr: Magnet):
        self._mag = corr

    def get(self) -> float:
        return np.arctan(self._mag.strength.get())

    def set(self, value: float):
        self._mag.strength.set(np.tan(value))

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return "rad"



# PyAML factory (construct AML objects from config files)
import fnmatch
import importlib
from threading import Lock

from pydantic import ValidationError

from ..common.element import Element
from ..common.exception import PyAMLConfigException


class BuildStrategy:
    def can_handle(self, module: object, config_dict: dict) -> bool:
        """Return True if this strategy can handle the module/config."""
        raise NotImplementedError

    def build(self, module: object, config_dict: dict):
        """Build the object according to custom logic."""
        raise NotImplementedError


class PyAMLFactory:
    """Singleton factory to build PyAML elements with future compatibility logic."""

    _instance = None
    _lock = Lock()

    def __new__(cls):
        """
        No matter how many times you call PyAMLFactory(),
        it will be created only once.
        """
        with cls._lock:
            if cls._instance is None:
                cls._instance = super().__new__(cls)
                cls._instance._elements = {}
                cls._instance._strategies = []
            return cls._instance

    def register_strategy(self, strategy: BuildStrategy):
        """Register a plugin-based strategy for object creation."""
        self._strategies.append(strategy)

    def remove_strategy(self, strategy: BuildStrategy):
        """Register a plugin-based strategy for object creation."""
        self._strategies.remove(strategy)

    def handle_validation_error(
        self, e, type_str: str, location_str: str, field_locations: dict
    ):
        # Handle pydantic errors
        globalMessage = ""
        for err in e.errors():
            msg = err["msg"]
            field = ""
            if len(err["loc"]) == 2:
                field, fieldIdx = err["loc"]
                message = f"'{field}.{fieldIdx}': {msg}"
            else:
                field = err["loc"][0]
                message = f"'{field}': {msg}"
            if field_locations and field in field_locations:
                file, line, col = field_locations[field]
                loc = f"{file} at line {line}, colum {col}"
                message += f" {loc}"
            globalMessage += message
            globalMessage += ", "
        # Discard pydantic stack trace
        raise PyAMLConfigException(
            f"{globalMessage} for object: '{type_str}' {location_str}"
        ) from None

    def build_object(self, d: dict, ignore_external: bool = False):
        """Build an object from the dict"""
        location = d.pop("__location__", None)
        field_locations = d.pop("__fieldlocations__", None)
        location_str = ""
        if location:
            file, line, col = location
            location_str = f"{file} at line {line}, column {col}."

        if not isinstance(d, dict):
            raise PyAMLConfigException(f"Unexpected object {str(d)} {location_str}")
        if "type" not in d:
            raise PyAMLConfigException(
                f"No type specified for {str(type(d))}:{str(d)} {location_str}"
            )
        type_str = d.pop("type")

        try:
            module = importlib.import_module(type_str)
        except ModuleNotFoundError as ex:
            if not ignore_external:
                # Discard module not found stack trace
                raise PyAMLConfigException(
                    "Module referenced in type cannot be found:"
                    + f"'{type_str}' {location_str}"
                ) from None
            else:
                return None

        # Try plugin strategies first
        for strategy in self._strategies:
            try:
                if strategy.can_handle(module, d):
                    obj = strategy.build(module, d)
                    self.register_element(obj)
                    return obj
            except Exception as e:
                raise PyAMLConfigException(
                    f"Custom strategy failed {location_str}"
                ) from e

        # Default loading strategy
        # Get the config object
        config_cls = getattr(module, "ConfigModel", None)
        if config_cls is None:
            raise PyAMLConfigException(
                f"ConfigModel class '{type_str}.ConfigModel' not found {location_str}"
            )

        # Get the class name
        cls_name = getattr(module, "PYAMLCLASS", None)
        if cls_name is None:
            raise PyAMLConfigException(
                f"PYAMLCLASS definition not found in '{type_str}' {location_str}"
            )

        try:
            # Validate the model
            cfg = config_cls.model_validate(d)
        except ValidationError as e:
            self.handle_validation_error(e, type_str, location_str, field_locations)

        # Construct and return the object
        elem_cls = getattr(module, cls_name, None)
        if elem_cls is None:
            raise PyAMLConfigException(f"Unknown element class '{type_str}.{cls_name}'")

        try:
            obj = elem_cls(cfg)
            self.register_element(obj)
        except Exception as e:
            raise PyAMLConfigException(
                f"{str(e)} when creating '{type_str}.{cls_name}' {location_str}"
            ) from e

        return obj

    def depth_first_build(self, d, ignore_external: bool):
        """
        Main factory function (Depth-first factory)

        Parameters
        ----------
        ignore_external: bool
            Ignore `module not found` and return None when an object cannot be created
        """

        if isinstance(d, list):
            # list can be a list of objects or a list of native types
            l = []
            for _index, e in enumerate(d):
                if isinstance(e, dict) or isinstance(e, list):
                    obj = self.depth_first_build(e, ignore_external)
                    l.append(obj)
                else:
                    l.append(e)
            return l

        elif isinstance(d, dict):
            for key, value in d.items():
                if not key == "__fieldlocations__":
                    if isinstance(value, dict) or isinstance(value, list):
                        obj = self.depth_first_build(value, ignore_external)
                        # Replace the inner dict by the object itself
                        d[key] = obj

            # We are now on leaf (no nested object), we can construct
            return self.build_object(d, ignore_external)

        raise PyAMLConfigException(
            "Unexpected element found. 'dict' or 'list' expected "
            "but got '{d.__class__.__name__}'"
        )

    def register_element(self, elt):
        if isinstance(elt, Element):
            name = elt.get_name()
            if name in self._elements:
                raise PyAMLConfigException(f"element {name} already defined")
            self._elements[name] = elt

    def get_element(self, name: str):
        if name not in self._elements:
            raise PyAMLConfigException(f"element {name} not defined")
        return self._elements[name]

    def get_elements_by_name(self, wildcard: str) -> list[Element]:
        return [e for k, e in self._elements.items() if fnmatch.fnmatch(k, wildcard)]

    def get_elements_by_type(self, type) -> list[Element]:
        return [e for k, e in self._elements.items() if isinstance(e, type)]

    def clear(self):
        self._elements.clear()


Factory = PyAMLFactory()



import numpy as np

from ..bpm.bpm import BPM
from ..common.abstract import ReadFloatArray
from ..control.deviceaccesslist import DeviceAccessList
from .element_array import ElementArray


class RWBPMPosition(ReadFloatArray):
    def __init__(self, name: str, bpms: list[BPM]):
        self.__bpms = bpms
        self.__name = name
        self.__aggregator: DeviceAccessList = None

    # Gets the values
    def get(self) -> np.array:
        if not self.__aggregator:
            return np.array([b.positions.get() for b in self.__bpms])
        else:
            return self.__aggregator.get().reshape(len(self.__bpms), 2)

    # Gets the unit of the values
    def unit(self) -> list[str]:
        return [b.positions.unit() for b in self.__bpms]

    # Set the aggregator (Control system only)
    def set_aggregator(self, agg: DeviceAccessList):
        self.__aggregator = agg


class RWBPMSinglePosition(ReadFloatArray):
    def __init__(self, name: str, bpms: list[BPM], idx: int):
        self.__bpms = bpms
        self.__name = name
        self.__idx = idx
        self.__aggregator: DeviceAccessList = None

    # Gets the values
    def get(self) -> np.array:
        if not self.__aggregator:
            return np.array([b.positions.get()[self.__idx] for b in self.__bpms])
        else:
            return self.__aggregator.get()

    # Gets the unit of the values
    def unit(self) -> list[str]:
        return [b.positions.unit() for b in self.__bpms]

    # Set the aggregator (Control system only)
    def set_aggregator(self, agg: DeviceAccessList):
        self.__aggregator = agg


class BPMArray(ElementArray):
    """
    Class that implements access to a BPM array

    Parameters
    ----------
    arrayName : str
        Array name
    bpms : list[BPM]
        BPM list, all elements must be attached to the same instance of
        either a Simulator or a ControlSystem.
    use_aggregator : bool
        Use aggregator to increase performance by using paralell
        access to underlying devices.

    Example
    -------

    An array can be retrieved from the configuration as in the following example::

        sr = Accelerator.load("acc.yaml")
        bpms = sr.design.get_bpms("BPMs")


    """

    def __init__(self, arrayName: str, bpms: list[BPM], use_aggregator=True):
        super().__init__(arrayName, bpms, use_aggregator)

        self.__hvpos = RWBPMPosition(arrayName, bpms)
        self.__hpos = RWBPMSinglePosition(arrayName, bpms, 0)
        self.__vpos = RWBPMSinglePosition(arrayName, bpms, 1)

        if use_aggregator and len(bpms) > 0:
            aggs = self.get_peer().create_bpm_aggregators(bpms)
            self.__hvpos.set_aggregator(aggs[0])
            self.__hpos.set_aggregator(aggs[1])
            self.__vpos.set_aggregator(aggs[2])

    @property
    def positions(self) -> RWBPMPosition:
        """
        Returns position of each bpm of this array
        """
        return self.__hvpos

    @property
    def h(self) -> RWBPMSinglePosition:
        """
        Returns horizontal position of each bpm of this array
        """
        return self.__hpos

    @property
    def v(self) -> RWBPMSinglePosition:
        """
        Returns vertical position of each bpm of this array
        """
        return self.__vpos



import numpy as np

from ..common.abstract import ReadWriteFloatArray
from ..common.exception import PyAMLException
from ..magnet.cfm_magnet import CombinedFunctionMagnet
from .element_array import ElementArray

# TODO handle aggregator for CFM


class RWMagnetStrengths(ReadWriteFloatArray):
    def __init__(self, name: str, magnets: list[CombinedFunctionMagnet]):
        self.__name = name
        self.__magnets = magnets
        self.__nb = sum(m.nb_multipole() for m in magnets)

    # Gets the values
    def get(self) -> np.array:
        r = np.zeros(self.__nb)
        idx = 0
        for m in self.__magnets:
            r[idx : idx + m.nb_multipole()] = m.strengths.get()
            idx += m.nb_multipole()
        return r

    # Sets the values
    def set(self, value: np.array):
        nvalue = np.ones(self.__nb) * value if isinstance(value, float) else value
        idx = 0
        for m in self.__magnets:
            m.strengths.set(nvalue[idx : idx + m.nb_multipole()])
            idx += m.nb_multipole()

    # Sets the values and waits that the read values reach their setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the values
    def unit(self) -> list[str]:
        r = []
        for m in self.__magnets:
            r.extend(m.strengths.unit())
        return r


class RWMagnetHardwares(ReadWriteFloatArray):
    def __init__(self, name: str, magnets: list[CombinedFunctionMagnet]):
        self.__name = name
        self.__magnets = magnets
        self.__nb = sum(m.nb_multipole() for m in magnets)

    # Gets the values
    def get(self) -> np.array:
        r = np.zeros(self.__nb)
        idx = 0
        for m in self.__magnets:
            r[idx : idx + m.nb_multipole()] = m.hardwares.get()
            idx += m.nb_multipole()
        return r

    # Sets the values
    def set(self, value: np.array):
        nvalue = np.ones(self.__nb) * value if isinstance(value, float) else value
        idx = 0
        for m in self.__magnets:
            m.hardwares.set(nvalue[idx : idx + m.nb_multipole()])
            idx += m.nb_multipole()

    # Sets the values and waits that the read values reach their setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the values
    def unit(self) -> list[str]:
        r = []
        for m in self.__magnets:
            r.extend(m.hardwares.unit())
        return r


class CombinedFunctionMagnetArray(ElementArray):
    """
    Class that implements access to a combined function magnet array

    Parameters
    ----------
    arrayName : str
        Array name
    magnets : list[Magnet]
        Magnet list, all elements must be attached to the same instance of
        either a Simulator or a ControlSystem.
    use_aggregator : bool
        Use aggregator to increase performance by using paralell
        access to underlying devices.
    """

    def __init__(
        self,
        arrayName: str,
        magnets: list[CombinedFunctionMagnet],
        use_aggregator=False,
    ):
        super().__init__(arrayName, magnets, use_aggregator)

        self.__rwstrengths = RWMagnetStrengths(arrayName, magnets)
        self.__rwhardwares = RWMagnetHardwares(arrayName, magnets)

        if use_aggregator:
            raise (
                PyAMLException(
                    "Aggregator not implemented for CombinedFunctionMagnetArray"
                )
            )

    @property
    def strengths(self) -> RWMagnetStrengths:
        """
        Give access to strength of each magnet of this array
        """
        return self.__rwstrengths

    @property
    def hardwares(self) -> RWMagnetHardwares:
        """
        Give access to hardware value of each magnet of this array
        """
        return self.__rwhardwares



"""
PyAML global constants
"""

# Horizontal kick sign convention for both kick an kick angle
HORIZONTAL_KICK_SIGN: float = -1.0

# Action identifier for callback in response matrix measurement
ACTION_APPLY: int = 0
ACTION_RESTORE: int = 1



from abc import ABCMeta, abstractmethod

import numpy as np
from numpy.typing import NDArray

from ..control.deviceaccess import DeviceAccess


class BPMModel(metaclass=ABCMeta):
    """
    Abstract class providing interface to accessing BPM positions, offsets,
    tilts.
    """

    @abstractmethod
    def get_pos_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles used for position reading

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        pass

    @abstractmethod
    def get_tilt_device(self) -> DeviceAccess | None:
        """
        Get device handle used for tilt access

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        pass

    @abstractmethod
    def get_offset_devices(self) -> list[DeviceAccess | None]:
        """
        Get device handles used for offset access

        Returns
        -------
        list[DeviceAccess]
            Array of DeviceAcess
        """
        pass



from pathlib import Path

from pyaml.accelerator import Accelerator
from pyaml.tuning_tools.orbit_response_matrix import ConfigModel as ORM_ConfigModel
from pyaml.tuning_tools.orbit_response_matrix import OrbitResponseMatrix

parent_folder = Path(__file__).parent
config_path = parent_folder.parent.parent.joinpath(
    "tests", "config", "EBSOrbit.yaml"
).resolve()
sr = Accelerator.load(config_path)
element_holder = sr.design

orm = OrbitResponseMatrix(
    cfg=ORM_ConfigModel(
        bpm_array_name="BPM",
        hcorr_array_name="HCorr",
        vcorr_array_name="VCorr",
        corrector_delta=1e-6,
    ),
    element_holder=element_holder,
)

orm.measure()
orm.save(parent_folder / Path("ideal_orm.json"))
orm.save(parent_folder / Path("ideal_orm.yaml"), with_type="yaml")
orm.save(parent_folder / Path("ideal_orm.npz"), with_type="npz")

ormdata = orm.get()



"""
Class providing polynom information
"""


class PolynomInfo:
    """
    Polynom information
    """

    def __init__(self, attName: str, index: int, sign: float = 1.0):
        """
        Construct a polynom information object

        Parameters
        ----------
        attName : str
            Name of the AT element attribute ('PolynomA' or 'PolynomB')
        index : int
            Polynomial coefficients index (sarting from 0)
        """
        self.attName = attName
        self.index = index
        self.sign = sign



import numpy as np
from pydantic import BaseModel, ConfigDict

from ..common.element import __pyaml_repr__
from ..common.exception import PyAMLException
from ..configuration.curve import Curve
from ..configuration.matrix import Matrix
from ..control.deviceaccess import DeviceAccess
from .model import MagnetModel

# Define the main class name for this module
PYAMLCLASS = "LinearCFMagnetModel"


class ConfigModel(BaseModel):
    """
    Configuration model for linear combined function magnet model

    Parameters
    ----------
    multipoles : list[str]
        List of supported functions: A0, B0, A1, B1, etc (i.e. [B0, A1, B2])
    curves : list[Curve]
        Excitation curves, 1 curve per function
    calibration_factors : list[float], optional
        Correction factor applied to curves, 1 factor per function.
        Default: ones
    calibration_offsets : list[float], optional
        Correction offset applied to curves, 1 offset per function.
        Default: zeros
    pseudo_factors : list[float], optional
        Factors applied to 'pseudo currents', 1 factor per function.
        Default: ones
    pseudo_offsets : list[float], optional
        Offsets applied to 'pseudo currents', 1 offset per function.
        Default: zeros
    powerconverters : list[DeviceAccess]
        List of power converter devices to apply currents (can be different
        from number of functions)
    matrix : Matrix, optional
        n x m matrix (n rows for n functions, m columns for m currents)
        to handle multipoles separation. Default: Identity
    units : list[str]
        List of strength units (i.e. ['rad', 'm-1', 'm-2'])
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    multipoles: list[str]
    curves: list[Curve]
    calibration_factors: list[float] = None
    calibration_offsets: list[float] = None
    pseudo_factors: list[float] = None
    pseudo_offsets: list[float] = None
    powerconverters: list[DeviceAccess | None]
    matrix: Matrix = None
    units: list[str]


class LinearCFMagnetModel(MagnetModel):
    """
    Class providing a simple linear model for combined function magnets. A matrix
    can handle separation of multipoles. A pseudo current is a linear combination
    of power supply currents associated to a single function.
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        self._brho = np.nan

        # Check config
        self.__nbFunction: int = len(cfg.multipoles)
        self.__nbPS: int = len(cfg.powerconverters)

        if cfg.calibration_factors is None:
            self.__calibration_factors = np.ones(self.__nbFunction)
        else:
            self.__calibration_factors = cfg.calibration_factors

        if cfg.calibration_offsets is None:
            self.__calibration_offsets = np.zeros(self.__nbFunction)
        else:
            self.__calibration_offsets = cfg.calibration_offsets

        if cfg.pseudo_factors is None:
            self.__pf = np.ones(self.__nbFunction)
        else:
            self.__pf = cfg.pseudo_factors

        if cfg.pseudo_offsets is None:
            self.__po = np.zeros(self.__nbFunction)
        else:
            self.__po = cfg.pseudo_factors

        self.__check_len(
            self.__calibration_factors, "calibration_factors", self.__nbFunction
        )
        self.__check_len(
            self.__calibration_offsets, "calibration_offsets", self.__nbFunction
        )
        self.__check_len(self.__pf, "pseudo_factors", self.__nbFunction)
        self.__check_len(self.__po, "pseudo_offsets", self.__nbFunction)
        self.__check_len(cfg.units, "units", self.__nbFunction)
        self.__check_len(cfg.curves, "curves", self.__nbFunction)

        if cfg.matrix is None:
            self.__matrix = np.identity(self.__nbFunction)
        else:
            self.__matrix = cfg.matrix.get_matrix()

        _s = np.shape(self.__matrix)

        if len(_s) != 2 or _s[0] != self.__nbFunction or _s[1] != self.__nbPS:
            raise PyAMLException(
                "matrix wrong dimension "
                f"({self.__nbFunction}x{self.__nbPS} expected but got {_s[0]}x{_s[1]})"
            )

        self.__curves = []
        self.__rcurves = []

        # Apply factor and offset
        for idx, c in enumerate(cfg.curves):
            self.__curves.append(c.get_curve())
            self.__curves[idx][:, 1] *= self.__calibration_factors[idx]
            self.__curves[idx][:, 1] += self.__calibration_offsets[idx]
            self.__rcurves.append(Curve.inverse(self.__curves[idx]))

        # Compute pseudo inverse
        self.__inv = np.linalg.pinv(self.__matrix)

    def __check_len(self, obj, name, expected_len):
        lgth = len(obj)
        if lgth != expected_len:
            raise PyAMLException(
                f"{name} does not have the expected "
                f"number of items ({expected_len} items expected but got {lgth})"
            )

    def compute_hardware_values(self, strengths: np.array) -> np.array:
        _pI = np.zeros(self.__nbFunction)
        for idx, c in enumerate(self.__rcurves):
            _pI[idx] = (
                self.__pf[idx]
                * np.interp(strengths[idx] * self._brho, c[:, 0], c[:, 1])
                + self.__po[idx]
            )
        _currents = np.matmul(self.__inv, _pI)
        return _currents

    def compute_strengths(self, currents: np.array) -> np.array:
        _strength = np.zeros(self.__nbFunction)
        _pI = np.matmul(self.__matrix, currents)
        for idx, c in enumerate(self.__curves):
            _strength[idx] = (
                np.interp(
                    (_pI[idx] - self.__po[idx]) / self.__pf[idx], c[:, 0], c[:, 1]
                )
                / self._brho
            )
        return _strength

    def get_strength_units(self) -> list[str]:
        return self._cfg.units

    def get_hardware_units(self) -> list[str]:
        return np.array([p.unit() for p in self._cfg.powerconverters])

    def get_devices(self) -> list[DeviceAccess]:
        return self._cfg.powerconverters

    def set_magnet_rigidity(self, brho: np.double):
        self._brho = brho

    def has_hardware(self) -> bool:
        return (self.__nbPS == self.__nbFunction) and np.allclose(
            self.__matrix, np.eye(self.__nbFunction)
        )

    def __repr__(self):
        return __pyaml_repr__(self)



import logging
from pathlib import Path
from typing import List, Optional

from pydantic import BaseModel, ConfigDict

from ..common.element_holder import ElementHolder
from ..external.pySC.pySC.apps import measure_dispersion
from ..external.pySC_interface import pySCInterface

logger = logging.getLogger(__name__)

PYAMLCLASS = "OrbitResponseMatrix"


class ConfigModel(BaseModel):
    """
    Configuration model for dispersion measurement

    Parameters
    ----------
    bpm_array_name : str
        BPM array name
    rf_plant_name : str
        RF plant name
    frequency_delta : float
        Frequency delta for measurement
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    bpm_array_name: str
    rf_plant_name: str
    frequency_delta: float


class Dispersion(object):
    def __init__(self, element_holder: ElementHolder, cfg: ConfigModel):
        self._cfg = cfg

        self.element_holder = element_holder
        self.bpm_array_name = cfg.bpm_array_name
        self.rf_plant_name = cfg.rf_plant_name
        self.frequency_delta = cfg.frequency_delta
        self.latest_measurement = None

    def measure(self):
        interface = pySCInterface(
            element_holder=self.element_holder,
            bpm_array_name=self.bpm_array_name,
            rf_plant_name=self.rf_plant_name,
        )

        generator = measure_dispersion(
            interface=interface,
            delta=self.frequency_delta,
            skip_save=True,
        )

        _, measurement = next(generator)
        for _, _ in generator:
            pass

        dispersion_data = measurement.dispersion_data
        # contains also pre-processed data

        # dispersion_data.output_names = self.element_holder.get_bpms(
        #     self.bpm_array_name
        # ).names()
        self.latest_measurement = dispersion_data.model_dump()

    def get(self):
        return self.latest_measurement



from pathlib import Path

import numpy as np

from pyaml.accelerator import Accelerator
from pyaml.external.pySC_interface import pySCInterface

config_path = (
    Path(__file__)
    .parent.parent.parent.joinpath("tests", "config", "EBSOrbit.yaml")
    .resolve()
)
sr = Accelerator.load(config_path)

live_h = sr.live.get_bpms("BPM").h.get()  # ok
design_h = sr.design.get_bpms("BPM").h.get()  # ok

interface = pySCInterface(element_holder=sr.design)

orbit_x, orbit_y = interface.get_orbit()

one_hcorr = interface.hcorr_names[10]

sp0 = interface.get(one_hcorr)
interface.set(one_hcorr, sp0 + 10e-6)

hcorr_setpoints = interface.get_many(interface.hcorr_names)
interface.set_many(hcorr_setpoints)

orbit_x2, orbit_y2 = interface.get_orbit()

print(np.std(orbit_x2 - orbit_x))



from pathlib import Path

import numpy as np
from pydantic import BaseModel, ConfigDict

from ..common.exception import PyAMLException
from ..configuration.fileloader import get_path
from .curve import Curve

# Define the main class name for this module
PYAMLCLASS = "CSVCurve"


class ConfigModel(BaseModel):
    """
    Configuration model for CSV curve

    Parameters
    ----------
    file : str
        CSV file that contains the curve (n rows, 2 columns)
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    file: str


class CSVCurve(Curve):
    """
    Class for load CSV (x,y) curve
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg

        # Load CSV curve
        path = get_path(cfg.file)
        try:
            self._curve = np.genfromtxt(path, delimiter=",", dtype=float, loose=False)
        except ValueError as e:
            raise PyAMLException(
                f"CSVCurve(file='{cfg.file}',dtype=float): {str(e)}"
            ) from None

        _s = np.shape(self._curve)
        if len(_s) != 2 or _s[1] != 2:
            raise PyAMLException(
                f"CSVCurve(file='{cfg.file}',dtype=float):"
                f"wrong shape (2,2) expected but got {str(_s)}"
            )

    def get_curve(self) -> np.array:
        return self._curve

    def __repr__(self):
        return repr(self._cfg).replace("ConfigModel", self.__class__.__name__)



from pyaml.magnet.hcorrector import HCorrector
from pyaml.magnet.octupole import Octupole
from pyaml.magnet.quadrupole import Quadrupole
from pyaml.magnet.sextupole import Sextupole
from pyaml.magnet.skewoctu import SkewOctu
from pyaml.magnet.skewquad import SkewQuad
from pyaml.magnet.skewsext import SkewSext
from pyaml.magnet.vcorrector import VCorrector

function_map: dict = {
    "B0": HCorrector,
    "A0": VCorrector,
    "B1": Quadrupole,
    "A1": SkewQuad,
    "B2": Sextupole,
    "A2": SkewSext,
    "B3": Octupole,
    "A3": SkewOctu,
}



import json
import logging
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from pyaml.accelerator import Accelerator
from pyaml.tuning_tools.orbit import ConfigModel as Orbit_ConfigModel
from pyaml.tuning_tools.orbit import Orbit

parent_folder = Path(__file__).parent
pyaml_folder = parent_folder.parent.parent
config_path = pyaml_folder.joinpath("tests/config/EBSOrbit.yaml").resolve()
sr = Accelerator.load(config_path)
ebs = sr.design

## get reference
ref_h, ref_v = ebs.get_bpms("BPM").positions.get().T
reference = np.concat((ref_h, ref_v))
########################################################


## generate some orbit
std_kick = 1e-6
hcorr = ebs.get_magnets("HCorr")
vcorr = ebs.get_magnets("VCorr")
bpms = ebs.get_bpms("BPM")

# mangle orbit
hcorr.strengths.set(
    hcorr.strengths.get() + std_kick * np.random.normal(size=len(hcorr))
)
vcorr.strengths.set(
    vcorr.strengths.get() + std_kick * np.random.normal(size=len(vcorr))
)

positions_bc = bpms.positions.get()
std_bc = np.std(positions_bc, axis=0)
print(
    "R.m.s. orbit before correction "
    f"H: {1e6 * std_bc[0]: .1f} µm, V: {1e6 * std_bc[1]: .1f} µm."
)
########################################################

## Correct the orbit
ebs.orbit.correct(reference=reference)
# ebs.orbit.correct(plane="H")
# ebs.orbit.correct(plane="V")
########################################################

## inspect orbit correction
positions_ac = bpms.positions.get()
std_ac = np.std(positions_ac, axis=0)
print(
    "R.m.s. orbit after correction H: "
    f"{1e6 * std_ac[0]: .1f} µm, V: {1e6 * std_ac[1]: .1f} µm,"
)

fig = plt.figure()
ax1 = fig.add_subplot(311)
ax2 = fig.add_subplot(312)
ax3 = fig.add_subplot(313)
ax1.plot(positions_bc[:, 0] * 1e6, label="Orbit before correction")
ax2.plot(positions_bc[:, 1] * 1e6, label="Orbit before correction")
ax1.plot(positions_ac[:, 0] * 1e6, label="Orbit after correction")
ax2.plot(positions_ac[:, 1] * 1e6, label="Orbit after correction")

ax3.plot(hcorr.strengths.get())
ax3.plot(vcorr.strengths.get())

ax1.set_ylabel("Horizontal pos. [μm]")
ax2.set_ylabel("Vertical pos. [μm]")
ax2.set_xlabel("BPM number")
ax3.set_ylabel("Strength (rad)")
ax3.set_xlabel("Steerer number")
fig.tight_layout()

plt.show()



from abc import ABCMeta, abstractmethod
from typing import Iterable

import at
from at import Lattice
from pydantic import BaseModel, ConfigDict

from pyaml import PyAMLException
from pyaml.common.element import Element


class LinkerConfigModel(BaseModel):
    """Base configuration model for linker definitions.

    This class defines the configuration structure used to instantiate
    a specific linking strategy. Each concrete implementation of a
    `LatticeElementsLinker` may define its own subclass extending this model
    to include additional configuration parameters.

    Attributes
    ----------
    model_config : ConfigDict
        Pydantic configuration allowing arbitrary field types and forbidding
        unexpected extra keys.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")


class LinkerIdentifier(metaclass=ABCMeta):
    """Abstract base class for identifiers used to match PyAML and PyAT elements.

    The identifier acts as an intermediate representation between the PyAML
    configuration and the PyAT lattice. Its exact structure depends on the
    linking strategy (e.g., family name, element index, or user-defined tag).

    Subclasses should define the fields and logic necessary to represent
    a unique reference to one or more PyAT elements.
    """

    pass


class LatticeElementsLinker(metaclass=ABCMeta):
    """Abstract base class defining the interface for PyAT–PyAML element linking.

    Implementations of this class define how PyAML elements are matched
    to PyAT elements based on a given linking strategy (e.g., by family name,
    by index, or by a custom attribute).

    Parameters
    ----------
    linker_config_model : LinkerConfigModel
        The configuration model for the linking strategy.

    Attributes
    ----------
    lattice : at.Lattice
        Reference to the PyAT lattice handled by this linker.
    """

    def __init__(self, linker_config_model: LinkerConfigModel):
        self.linker_config_model = linker_config_model
        self.lattice: Lattice = None

    def set_lattice(self, lattice: Lattice):
        self.lattice = lattice

    @abstractmethod
    def _test_at_element(
        self, identifier: LinkerIdentifier, element: at.Element
    ) -> bool:
        pass

    @abstractmethod
    def get_element_identifier(self, element: Element) -> LinkerIdentifier:
        pass

    def _iter_matches(self, identifier: LinkerIdentifier) -> Iterable[at.Element]:
        """Yield all elements in the lattice whose matches the identifier."""
        for elem in self.lattice:
            if self._test_at_element(identifier, elem):
                yield elem

    def get_at_elements(
        self, element_id: LinkerIdentifier | list[LinkerIdentifier]
    ) -> list[at.Element]:
        """Return a list of PyAT elements matching the given identifiers.

        This method should resolve one or multiple PyAML identifiers
        into their corresponding PyAT elements according to the specific
        linking strategy implemented.

        Parameters
        ----------
        element_id : LinkerIdentifier or list of LinkerIdentifier
            One or several identifiers describing which PyAT elements
            to retrieve.

        Returns
        -------
        list of at.Element
            The list of matching PyAT elements found in the lattice.

        Raises
        ------
        PyAMLException
            If no element matches the given identifier(s).
        """
        if isinstance(element_id, LinkerIdentifier):
            identifiers = [element_id]
        else:
            identifiers = element_id

        results: list[at.Element] = []
        for ident in identifiers:
            results.extend(self._iter_matches(ident))

        if not results:
            raise PyAMLException(
                f"No PyAT elements found for identifier(s): "
                f"{', '.join(i.__repr__() for i in identifiers)}"
            )
        return results

    def get_at_element(self, element_id: LinkerIdentifier) -> at.Element:
        """Return a single PyAT element matching the given identifier.

        Parameters
        ----------
        element_id : LinkerIdentifier
            Identifier describing the PyAT element to retrieve.

        Returns
        -------
        at.Element
            The PyAT element matching the identifier.

        Raises
        ------
        PyAMLException
            If no element matches the identifier.
        """
        for elem in self._iter_matches(element_id):
            return elem
        raise PyAMLException(
            f"No PyAT element found for FamName: {element_id.__repr__()}"
        )



from abc import ABCMeta, abstractmethod

from ..bpm.bpm import BPM
from ..common.abstract import RWMapper
from ..common.abstract_aggregator import ScalarAggregator
from ..common.element import Element
from ..common.element_holder import ElementHolder
from ..configuration.factory import Factory
from ..control.abstract_impl import (
    CSScalarAggregator,
    CSStrengthScalarAggregator,
    RBetatronTuneArray,
    RBpmArray,
    RWBpmOffsetArray,
    RWBpmTiltScalar,
    RWHardwareArray,
    RWHardwareScalar,
    RWRFFrequencyScalar,
    RWRFPhaseScalar,
    RWRFVoltageScalar,
    RWStrengthArray,
    RWStrengthScalar,
)
from ..diagnostics.tune_monitor import BetatronTuneMonitor
from ..magnet.cfm_magnet import CombinedFunctionMagnet
from ..magnet.magnet import Magnet
from ..magnet.serialized_magnet import SerializedMagnets
from ..rf.rf_plant import RFPlant, RWTotalVoltage
from ..rf.rf_transmitter import RFTransmitter
from ..tuning_tools.orbit import Orbit
from ..tuning_tools.tune import Tune
from .deviceaccess import DeviceAccess


class ControlSystem(ElementHolder, metaclass=ABCMeta):
    """
    Abstract class providing access to a control system float variable
    """

    def __init__(self):
        ElementHolder.__init__(self)

    @abstractmethod
    def attach(self, dev: list[DeviceAccess]) -> list[DeviceAccess]:
        """Return new instances of DeviceAccess objects
        coming from configuration attached to this CS"""
        pass

    @abstractmethod
    def name(self) -> str:
        """Return control system name (i.e. live)"""
        pass

    @abstractmethod
    def scalar_aggregator(self) -> str | None:
        """Returns the module name used for handling aggregator of DeviceAccess"""
        return None

    @abstractmethod
    def vector_aggregator(self) -> str | None:
        """Returns the module name used for handling aggregator of DeviceVectorAccess"""
        return None

    def create_scalar_aggregator(self) -> ScalarAggregator:
        mod = self.scalar_aggregator()
        agg = Factory.build_object({"type": mod}) if mod is not None else None
        return CSScalarAggregator(agg)

    def create_magnet_strength_aggregator(
        self, magnets: list[Magnet]
    ) -> ScalarAggregator:
        agg = CSStrengthScalarAggregator(self.create_scalar_aggregator())
        for m in magnets:
            devs = self.attach(m.model.get_devices())
            agg.add_magnet(m, devs)
        return agg

    def create_magnet_hardware_aggregator(
        self, magnets: list[Magnet]
    ) -> ScalarAggregator:
        """When working in hardware space, 1 single power
        supply device per multipolar strength is required
        """
        agg = self.create_scalar_aggregator()
        for m in magnets:
            if not m.model.has_hardware():
                return None
            psIndex = m.hardware.index() if isinstance(m.hardware, RWMapper) else 0
            agg.add_devices(self.attach([m.model.get_devices()[psIndex]])[0])
        return agg

    def create_bpm_aggregators(self, bpms: list[BPM]) -> list[ScalarAggregator]:
        agg = self.create_scalar_aggregator()
        aggh = self.create_scalar_aggregator()
        aggv = self.create_scalar_aggregator()
        for b in bpms:
            devs = self.attach(b.model.get_pos_devices())
            agg.add_devices(devs)
            aggh.add_devices(devs[0])
            aggv.add_devices(devs[1])
        return [agg, aggh, aggv]

    def set_energy(self, E: float):
        """
        Sets the energy on magnets belonging to this control system

        Parameters
        ----------
        E : float
            Energy in eV
        """
        # Needed by energy dependant element (i.e. magnet coil current calculation)
        for m in self.get_all_elements():
            m.set_energy(E)

    def fill_device(self, elements: list[Element]):
        """
        Fill device of this control system with Element
        coming from the configuration file

        Parameters
        ----------
        elements : list[Element]
            List of elements coming from the configuration
            file to attach to this control system
        """
        for e in elements:
            if isinstance(e, Magnet):
                dev = self.attach(e.model.get_devices())[0]
                current = (
                    RWHardwareScalar(e.model, dev) if e.model.has_hardware() else None
                )
                strength = (
                    RWStrengthScalar(e.model, dev) if e.model.has_physics() else None
                )
                # Create a unique ref for this control system
                m = e.attach(self, strength, current)
                self.add_magnet(m)

            elif isinstance(e, CombinedFunctionMagnet):
                devs = self.attach(e.model.get_devices())
                currents = RWHardwareArray(e.model, devs)
                strengths = RWStrengthArray(e.model, devs)
                # Create unique refs the cfm and
                # each of its function for this control system
                ms = e.attach(self, strengths, currents)
                self.add_cfm_magnet(ms[0])
                for m in ms[1:]:
                    self.add_magnet(m)

            elif isinstance(e, SerializedMagnets):
                devs = self.attach(e.model.get_devices())
                currents = []
                strengths = []
                # Create unique refs the series and each of its function for this control system
                for i in range(e.get_nb_magnets()):
                    current = (
                        RWHardwareScalar(e.model.get_sub_model(i), devs[i])
                        if e.model.has_hardware()
                        else None
                    )
                    strength = (
                        RWStrengthScalar(e.model.get_sub_model(i), devs[i])
                        if e.model.has_physics()
                        else None
                    )
                    currents.append(current)
                    strengths.append(strength)
                ms = e.attach(self, strengths, currents)
                self.add_serialized_magnet(ms[0])
                for m in ms[1:]:
                    self.add_magnet(m)

            elif isinstance(e, BPM):
                tiltDev = self.attach([e.model.get_tilt_device()])[0]
                offsetsDevs = self.attach(e.model.get_offset_devices())
                posDevs = self.attach(e.model.get_pos_devices())
                tilt = RWBpmTiltScalar(e.model, tiltDev)
                offsets = RWBpmOffsetArray(e.model, offsetsDevs)
                positions = RBpmArray(e.model, posDevs)
                e = e.attach(self, positions, offsets, tilt)
                self.add_bpm(e)

            elif isinstance(e, RFPlant):
                attachedTrans: list[RFTransmitter] = []
                if e._cfg.transmitters:
                    for t in e._cfg.transmitters:
                        vDev = self.attach([t._cfg.voltage])[0]
                        pDev = self.attach([t._cfg.phase])[0]
                        voltage = RWRFVoltageScalar(t, vDev)
                        phase = RWRFPhaseScalar(t, pDev)
                        nt = t.attach(self, voltage, phase)
                        self.add_rf_transnmitter(nt)
                        attachedTrans.append(nt)

                fDev = self.attach([e._cfg.masterclock])[0]
                frequency = RWRFFrequencyScalar(e, fDev)
                voltage = RWTotalVoltage(attachedTrans) if e._cfg.transmitters else None
                ne = e.attach(self, frequency, voltage)
                self.add_rf_plant(ne)

            elif isinstance(e, BetatronTuneMonitor):
                tuneDevs = self.attach([e._cfg.tune_h, e._cfg.tune_v])
                betatron_tune = RBetatronTuneArray(e, tuneDevs)
                e = e.attach(self, betatron_tune)
                self.add_betatron_tune_monitor(e)

            elif isinstance(e, Tune):
                self.add_tune_tuning(e.attach(self))

            elif isinstance(e, Orbit):
                self.add_orbit_tuning(e.attach(self))



from typing import TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from .exception import PyAMLException

if TYPE_CHECKING:
    from ..common.element_holder import ElementHolder


def __pyaml_repr__(obj):
    """
    Returns a string representation of a pyaml object
    """
    if hasattr(obj, "_cfg"):
        if isinstance(obj, Element):
            return repr(obj._cfg).replace(
                "ConfigModel(",
                obj.__class__.__name__ + "(peer='" + obj.get_peer() + "', ",
            )
        else:
            # no peer
            return repr(obj._cfg).replace("ConfigModel", obj.__class__.__name__)
    else:
        # Object is not yet fully constructed
        if isinstance(obj, Element):
            return f"{obj.__class__.__name__}: {obj.get_name()}"
        else:
            return f"{obj.__class__.__name__}"


class ElementConfigModel(BaseModel):
    """
    Base class for element configuration.

    Parameters
    ----------
    name : str
        The name of the PyAML element.
    lattice_names : str or None, optional
        The name(s) of the associated element(s) in the lattice. By default,
        the PyAML element name is used. lattice_name accept the following
        syntax:
        - list(name,[name]) : Element names
        - [name]@idx[,idx] : Element indices in the subset formed by name.
        - [name]#start_idx..end_idx : Element range in the subset formed by name.
        In the above syntax, if the name is not specficied, the whole set
        of lattice element is used for indexing.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    name: str
    lattice_names: str | None = None


class Element(object):
    """
    Class providing access to one element of a physical or simulated lattice

    Attributes:
      name: str
        The unique name identifying the element in the configuration file
    """

    def __init__(self, name: str):
        self.__name: str = name
        self._peer: "ElementHolder" = None  # Peer: ControlSystem, Simulator

    def get_name(self) -> str:
        """
        Returns the name of the element
        """
        return self.__name

    def get_lattice_names(self) -> str:
        """
        Returns the name of associated lattice element(s)
        """
        if not hasattr(self, "_cfg"):
            return self.__name
        else:
            return self._cfg.lattice_names

    def set_energy(self, E: float):
        """
        Set the instrument energy on this element
        """
        pass

    def check_peer(self):
        """
        Throws an exception if the element is not attacched
        to a simulator or to a control system
        """
        if self._peer is None:
            raise PyAMLException(
                f"{str(self)} is not attachedto a control system or the a simulator"
            )

    def get_peer(self) -> str:
        """
        Returns a string representation of peer simulator or control system
        """
        return (
            "None"
            if self._peer is None
            else f"{self._peer.__class__.__name__}:{self._peer.name()}"
        )

    def post_init(self):
        """
        Method triggered after all initialisations are done
        """
        pass

    def __repr__(self):
        return __pyaml_repr__(self)



import numpy as np
from numpy import double
from numpy.typing import NDArray

from ..bpm.bpm_model import BPMModel
from ..common import abstract
from ..common.abstract_aggregator import ScalarAggregator
from ..control.deviceaccess import DeviceAccess
from ..control.deviceaccesslist import DeviceAccessList
from ..magnet.magnet import Magnet
from ..magnet.model import MagnetModel
from ..rf.rf_plant import RFPlant
from ..rf.rf_transmitter import RFTransmitter

# ------------------------------------------------------------------------------


class CSScalarAggregator(ScalarAggregator):
    """
    Basic control system aggregator for a list of scalar values
    """

    def __init__(self, devs: DeviceAccessList):
        self._devs = devs

    def add_devices(self, devices: DeviceAccess | list[DeviceAccess]):
        self._devs.add_devices(devices)

    def set(self, value: NDArray[np.float64]):
        self._devs.set(value)

    def set_and_wait(self, value: NDArray[np.float64]):
        self._devs.set_and_wait(value)

    def get(self) -> NDArray[np.float64]:
        return self._devs.get()

    def readback(self) -> np.array:
        return self._devs.readback()

    def unit(self) -> str:
        return self._devs.unit()

    def nb_device(self) -> int:
        return self._devs.__len__()


# ------------------------------------------------------------------------------


class CSStrengthScalarAggregator(CSScalarAggregator):
    """
    Control system aggregator for a list of magnet strengths.
    This aggregator is in charge of computing hardware setpoints
    and applying them without overlap.
    When virtual magnets exported from combined function mangets are present (RWMapper),
    the aggregator prevents to apply several times the same power supply setpoint.
    """

    def __init__(self, peer: CSScalarAggregator):
        CSScalarAggregator.__init__(self, peer._devs)
        self.__models: list[MagnetModel] = []  # List of magnet model
        self.__modelToMagnet: list[list[tuple[int, int]]] = []  # strengths indexing
        self.__nbMagnet = 0  # Number of magnet strengths

    def add_magnet(self, magnet: Magnet, devs: list[DeviceAccess]):
        # Incoming magnet can be a magnet exported from
        # a CombinedFunctionMagnet or simple magnet.
        # All magnets exported from a same CombinedFunctionMagnet share the same model
        # TODO: check that strength is supported (m.strength may be None)
        strengthIndex = (
            magnet.strength.index()
            if isinstance(magnet.strength, abstract.RWMapper)
            else 0
        )
        if magnet.model not in self.__models:
            index = len(self.__models)
            self.__models.append(magnet.model)
            self.__modelToMagnet.append([(self.__nbMagnet, strengthIndex)])
            self._devs.add_devices(devs)
        else:
            index = self.__models.index(magnet.model)
            self.__modelToMagnet[index].append((self.__nbMagnet, strengthIndex))
        self.__nbMagnet += 1

    def set(self, value: NDArray[np.float64]):
        allHardwareValues = self._devs.get()  # Read all hardware setpoints
        newHardwareValues = np.zeros(self.nb_device())
        hardwareIndex = 0
        for modelIndex, model in enumerate(self.__models):
            nbDev = len(model.get_devices())
            mStrengths = model.compute_strengths(
                allHardwareValues[hardwareIndex : hardwareIndex + nbDev]
            )
            for valueIdx, strengthIdx in self.__modelToMagnet[modelIndex]:
                mStrengths[strengthIdx] = value[valueIdx]
            newHardwareValues[hardwareIndex : hardwareIndex + nbDev] = (
                model.compute_hardware_values(mStrengths)
            )
            hardwareIndex += nbDev
        self._devs.set(newHardwareValues)

    def set_and_wait(self, value: NDArray[np.float64]):
        raise NotImplementedError("Not implemented yet.")

    def get(self) -> NDArray[np.float64]:
        allHardwareValues = self._devs.get()  # Read all hardware setpoints
        allStrength = np.zeros(self.__nbMagnet)
        hardwareIndex = 0
        for modelIndex, model in enumerate(self.__models):
            nbDev = len(model.get_devices())
            mStrengths = model.compute_strengths(
                allHardwareValues[hardwareIndex : hardwareIndex + nbDev]
            )
            for valueIdx, strengthIdx in self.__modelToMagnet[modelIndex]:
                allStrength[valueIdx] = mStrengths[strengthIdx]
            hardwareIndex += nbDev
        return allStrength

    def readback(self) -> np.array:
        allHardwareValues = self._devs.readback()  # Read all hardware readback
        allStrength = np.zeros(self.__nbMagnet)
        hardwareIndex = 0
        for modelIndex, model in enumerate(self.__models):
            nbDev = len(model.get_devices())
            mStrengths = model.compute_strengths(
                allHardwareValues[hardwareIndex : hardwareIndex + nbDev]
            )
            for valueIdx, strengthIdx in self.__modelToMagnet[modelIndex]:
                allStrength[valueIdx] = mStrengths[strengthIdx]
            hardwareIndex += nbDev
        return allStrength

    def unit(self) -> str:
        return self._devs.unit()


# ------------------------------------------------------------------------------


class RWHardwareScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a magnet
    of a control system (in hardware units)
    """

    def __init__(self, model: MagnetModel, dev: DeviceAccess):
        self.__model = model
        self.__dev = dev

    def get(self) -> float:
        return self.__dev.get()

    def set(self, value: float):
        self.__dev.set(value)

    def set_and_wait(self, value: double):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self.__model.get_hardware_units()[0]

    def set_magnet_rigidity(self, brho: np.double):
        self.__model.set_magnet_rigidity(brho)


# ------------------------------------------------------------------------------


class RWStrengthScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to a strength of a control system
    """

    def __init__(self, model: MagnetModel, dev: DeviceAccess):
        self.__model = model
        self.__dev = dev

    # Gets the value
    def get(self) -> float:
        current = self.__dev.get()
        return self.__model.compute_strengths([current])[0]

    # Sets the value
    def set(self, value: float):
        current = self.__model.compute_hardware_values([value])[0]
        self.__dev.set(current)

    # Sets the value and wait that the read value reach the setpoint
    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return self.__model.get_strength_units()[0]

    def set_magnet_rigidity(self, brho: np.double):
        self.__model.set_magnet_rigidity(brho)


# ------------------------------------------------------------------------------


class RWHardwareArray(abstract.ReadWriteFloatArray):
    """
    Class providing read write access to a magnet array
    of a control system (in hardware units)
    """

    def __init__(self, model: MagnetModel, devs: list[DeviceAccess]):
        self.__model = model
        self.__devs = devs

    # Gets the value
    def get(self) -> np.array:
        return np.array([p.get() for p in self.__devs])

    # Sets the value
    def set(self, value: np.array):
        for idx, p in enumerate(self.__devs):
            p.set(value[idx])

    # Sets the value and waits that the read value reach the setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> list[str]:
        return self.__model.get_hardware_units()


# ------------------------------------------------------------------------------


class RWStrengthArray(abstract.ReadWriteFloatArray):
    """
    Class providing read write access to magnet strengths of a control system
    """

    def __init__(self, model: MagnetModel, devs: list[DeviceAccess]):
        self.__model = model
        self.__devs = devs

    # Gets the value
    def get(self) -> np.array:
        r = np.array([p.get() for p in self.__devs])
        str = self.__model.compute_strengths(r)
        return str

    # Sets the value
    def set(self, value: np.array):
        cur = self.__model.compute_hardware_values(value)
        for idx, p in enumerate(self.__devs):
            p.set(cur[idx])

    # Sets the value and waits that the read value reach the setpoint
    def set_and_wait(self, value: np.array):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> list[str]:
        return self.__model.get_strength_units()


# ------------------------------------------------------------------------------


class RBpmArray(abstract.ReadFloatArray):
    """
    Class providing read access to a BPM array of a control system
    """

    def __init__(self, model: BPMModel, devs: list[DeviceAccess]):
        self.__model = model
        self.__devs = devs

    # Gets the values
    def get(self) -> np.array:
        return np.array([self.__devs[0].get(), self.__devs[1].get()])

    # Gets the unit of the value Assume that x and y has the same unit
    def unit(self) -> str:
        return self.__model.get_pos_devices()[0].unit()


# ------------------------------------------------------------------------------


class RWBpmTiltScalar(abstract.ReadFloatScalar):
    """
    Class providing read access to a BPM tilt of a control system
    """

    def __init__(self, model: BPMModel, dev: DeviceAccess):
        self.__model = model
        self.__dev = dev

    # Gets the value
    def get(self) -> float:
        return self.__dev.get()

    def set(self, value: float):
        self.__dev.set(value)

    def set_and_wait(self, value: NDArray[np.float64]):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return self.__model.get_tilt_device().unit()


# ------------------------------------------------------------------------------


class RWBpmOffsetArray(abstract.ReadWriteFloatArray):
    """
    Class providing read write access to a BPM offset of a control system
    """

    def __init__(self, model: BPMModel, devs: list[DeviceAccess]):
        self.__model = model
        self.__devs = devs

    # Gets the value
    def get(self) -> NDArray[np.float64]:
        return np.array([self.__devs[0].get(), self.__devs[1].get()])

    # Sets the value
    def set(self, value: NDArray[np.float64]):
        self.__devs[0].set(value[0])
        self.__devs[1].set(value[1])

    def set_and_wait(self, value: NDArray[np.float64]):
        raise NotImplementedError("Not implemented yet.")

    # Gets the unit of the value
    def unit(self) -> str:
        return self.__model.get_offset_devices()[0].unit()


# ------------------------------------------------------------------------------


class RWRFVoltageScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to cavity voltage
    for a transmitter of a control system.
    """

    def __init__(self, transmitter: RFTransmitter, dev: DeviceAccess):
        self.__transmitter = transmitter
        self.__dev = dev

    def get(self) -> float:
        return self.__dev.get()

    def set(self, value: float):
        self.__dev.set(value)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self.__transmitter._cfg.voltage.unit()


# ------------------------------------------------------------------------------


class RWRFPhaseScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to cavity phase
    for a transmitter of a control system.
    """

    def __init__(self, transmitter: RFTransmitter, dev: DeviceAccess):
        self.__transmitter = transmitter
        self.__dev = dev

    def get(self) -> float:
        return self.__dev.get()

    def set(self, value: float):
        self.__dev.set(value)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self.__transmitter._cfg.phase.unit()


# ------------------------------------------------------------------------------


class RWRFFrequencyScalar(abstract.ReadWriteFloatScalar):
    """
    Class providing read write access to RF frequency of a control system.
    """

    def __init__(self, rf: RFPlant, dev: DeviceAccess):
        self.__rf = rf
        self.__dev = dev

    def get(self) -> float:
        # Serialized cavity has the same frequency
        return self.__dev.get()

    def set(self, value: float):
        self.__dev.set(value)

    def set_and_wait(self, value: float):
        raise NotImplementedError("Not implemented yet.")

    def unit(self) -> str:
        return self.__rf._cfg.masterclock.unit()


# ------------------------------------------------------------------------------


class RBetatronTuneArray(abstract.ReadFloatArray):
    """
    Class providing read write access to betatron tune of a control system.
    """

    def __init__(self, tune_monitor, devs: list[DeviceAccess]):
        self.__tune_monitor = tune_monitor
        self.__devs = devs

    def get(self) -> NDArray:
        # Return horizontal and vertical betatron tunes as a NumPy array
        return np.array(
            [
                self.__devs[0].get(),
                self.__devs[1].get(),
            ]
        )

    def unit(self) -> str:
        return self.__tune_monitor._cfg.tune_v.unit()



from ..common.element_holder import ElementHolder
from .array import ArrayConfig, ArrayConfigModel

# Define the main class name for this module
PYAMLCLASS = "Element"


class ConfigModel(ArrayConfigModel): ...


class Element(ArrayConfig):
    """
    Element array confirguration

    Example
    -------

    An element array configuration can also be created by code using
    the following example::

        from pyaml.arrays.element import Element,ConfigModel as ElementArrayConfigModel
        elemArray = Element(
                      ElementArrayConfigModel(name="MyArray", elements=["elt1","elt2"])
                    )


    """

    def __init__(self, cfg: ArrayConfigModel):
        super().__init__(cfg)

    def fill_array(self, holder: ElementHolder):
        holder.fill_element_array(self._cfg.name, self._cfg.elements)



from pathlib import Path

import matplotlib.pyplot as plt

from pyaml.accelerator import Accelerator
from pyaml.tuning_tools.dispersion import ConfigModel as Disp_ConfigModel
from pyaml.tuning_tools.dispersion import Dispersion

parent_folder = Path(__file__).parent
config_path = parent_folder.parent.parent.joinpath(
    "tests", "config", "EBSOrbit.yaml"
).resolve()
sr = Accelerator.load(config_path)
element_holder = sr.design

dispersion = Dispersion(
    cfg=Disp_ConfigModel(
        bpm_array_name="BPM",
        rf_plant_name="RF",
        frequency_delta=10,
    ),
    element_holder=element_holder,
)

dispersion.measure()
dispersion_data = dispersion.get()

fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(dispersion_data["frequency_response_x"])
plt.show()



import numpy as np
from pydantic import BaseModel, ConfigDict

from .matrix import Matrix

# Define the main class name for this module
PYAMLCLASS = "InlineMatrix"


class ConfigModel(BaseModel):
    """
    Configuration model for inline matrix

    Parameters
    ----------
    mat : list[list[float]]
        The matrix
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    mat: list[list[float]]


class InlineMatrix(Matrix):
    """
    Class for loading CSV matrix
    """

    def __init__(self, cfg: ConfigModel):
        self._cfg = cfg
        # Load the matrix
        self._mat = np.array(self._cfg.mat)

    def get_matrix(self) -> np.array:
        return self._mat

    def __repr__(self):
        return repr(self._cfg).replace("ConfigModel", self.__class__.__name__)



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Literal, Optional

try:
    from typing import Self  # Python 3.11+
except ImportError:
    from typing_extensions import Self  # Python 3.10 and earlier


from pydantic import BaseModel, ConfigDict

if TYPE_CHECKING:
    from ..common.element_holder import ElementHolder
# from ..external.pySC.pySC import ResponseMatrix as pySC_ResponseMatrix
from ..arrays.magnet_array import MagnetArray
from ..common.element import Element, ElementConfigModel
from ..common.exception import PyAMLException
from ..external.pySC.pySC import ResponseMatrix as pySC_ResponseMatrix
from ..external.pySC.pySC.apps import orbit_correction
from ..external.pySC_interface import pySCInterface
from .response_matrix import ResponseMatrix

logger = logging.getLogger(__name__)
logging.getLogger("pyaml.external.pySC").setLevel(logging.WARNING)

PYAMLCLASS = "Orbit"


class ConfigModel(ElementConfigModel):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    bpm_array_name: str
    hcorr_array_name: str
    vcorr_array_name: str
    singular_values: int
    response_matrix: Optional[ResponseMatrix]


class Orbit(Element):
    def __init__(self, cfg: ConfigModel):
        super().__init__(cfg.name)
        self._cfg = cfg

        self.bpm_array_name = cfg.bpm_array_name
        self.hcorr_array_name = cfg.hcorr_array_name
        self.vcorr_array_name = cfg.vcorr_array_name
        self.singular_values = cfg.singular_values
        self.response_matrix = pySC_ResponseMatrix.model_validate(
            cfg.response_matrix._cfg.model_dump()
        )
        self._hcorr: MagnetArray = None
        self._vcorr: MagnetArray = None
        self._hvcorr: MagnetArray = None

    def correct(
        self,
        reference=None,
        gain: float = 1.0,
        plane: Optional[Literal["H", "V"]] = None,
    ):
        interface = pySCInterface(
            element_holder=self._peer,
            bpm_array_name=self.bpm_array_name,
        )

        if plane is None or plane == "H":
            trims_h = orbit_correction(
                interface=interface,
                response_matrix=self.response_matrix,
                method="svd_values",
                parameter=self.singular_values,
                zerosum=True,
                apply=False,
                plane="H",
                reference=reference,
            )

        if plane is None or plane == "V":
            trims_v = orbit_correction(
                interface=interface,
                response_matrix=self.response_matrix,
                method="svd_values",
                parameter=self.singular_values,
                zerosum=False,
                apply=False,
                plane="V",
                reference=reference,
            )

        if plane is None:
            trims = {**trims_h, **trims_v}
            corr_array = self._hvcorr
        elif plane == "H":
            trims = trims_h
            corr_array = self._hcorr
        elif plane == "V":
            trims = trims_v
            corr_array = self._vcorr

        corrector_names = corr_array.names()
        corrector_to_index = {name: idx for idx, name in enumerate(corrector_names)}
        data_to_send = corr_array.strengths.get()
        for name in trims.keys():
            idx = corrector_to_index.get(name, None)
            if idx is None:
                raise PyAMLException(
                    f"Corrector {name} not found in the magnet array for orbit corr. "
                    "Possible inconcistency between corrector arrays and "
                    "response matrix."
                )
            data_to_send[idx] += trims[name] * gain

        corr_array.strengths.set(data_to_send)
        return

    def post_init(self):
        self._hcorr = self._peer.get_magnets(self._cfg.hcorr_array_name)
        self._vcorr = self._peer.get_magnets(self._cfg.vcorr_array_name)
        hvElts = []
        hvElts.extend(self._hcorr)
        hvElts.extend(self._vcorr)
        self._hvcorr = MagnetArray("", hvElts)

    def attach(self, peer: "ElementHolder") -> Self:
        """
        Create a new reference to attach this Orbit object to a simulator
        or a control system.
        """
        obj = self.__class__(self._cfg)
        obj._peer = peer
        return obj



import fnmatch
import importlib

from ..bpm.bpm import BPM
from ..common.element import Element
from ..common.exception import PyAMLException
from ..magnet.cfm_magnet import CombinedFunctionMagnet
from ..magnet.magnet import Magnet


class ElementArray(list[Element]):
    """
    Class that implements access to an element array

    Parameters
    ----------
    arrayName : str
        Array name
    elements : list[Element]
        Element list, all elements must be attached to the same instance of
        either a Simulator or a ControlSystem.
    use_aggregator : bool
        Use aggregator to increase performance by using paralell
        access to underlying devices.

    Example
    -------

    An array can be retrieved from the configuration as in the following example::

        sr = Accelerator.load("acc.yaml")
        elts = sr.design.get_elemens("QuadForTune")

    """

    def __init__(self, arrayName: str, elements: list[Element], use_aggregator=True):
        super().__init__(i for i in elements)
        self.__name = arrayName
        if len(elements) > 0:
            self.__peer = self[0]._peer if len(self) > 0 else None
            self.__use_aggretator = use_aggregator
            if self.__peer is None or any([m._peer != self.__peer for m in self]):
                raise PyAMLException(
                    f"{self.__class__.__name__} {self.get_name()}: "
                    "All elements must be attached to the same instance "
                    "of either a Simulator or a ControlSystem"
                )

    def get_peer(self):
        """
        Returns the peer (Simulator or ControlSystem) of an element list
        """
        return self.__peer

    def get_name(self) -> str:
        """
        Returns the array name
        """
        return self.__name

    def names(self) -> list[str]:
        """
        Returns the element names
        """
        return [e.get_name() for e in self]

    def __create_array(self, arrName: str, eltType: type, elements: list):
        if len(elements) == 0:
            return []

        if issubclass(eltType, Magnet):
            m = importlib.import_module("pyaml.arrays.magnet_array")
            arrayClass = getattr(m, "MagnetArray", None)
            return arrayClass("", elements, self.__use_aggretator)
        elif issubclass(eltType, BPM):
            m = importlib.import_module("pyaml.arrays.bpm_array")
            arrayClass = getattr(m, "BPMArray", None)
            return arrayClass("", elements, self.__use_aggretator)
        elif issubclass(eltType, CombinedFunctionMagnet):
            m = importlib.import_module("pyaml.arrays.cfm_magnet_array")
            arrayClass = getattr(m, "CombinedFunctionMagnetArray", None)
            return arrayClass("", elements, self.__use_aggretator)
        elif issubclass(eltType, Element):
            return ElementArray("", elements, self.__use_aggretator)
        else:
            raise PyAMLException(f"Unsupported sliced array for type {str(eltType)}")

    def __eval_field(self, attName: str, e: Element) -> str:
        funcName = "get_" + attName
        func = getattr(e, funcName, None)
        return func() if func is not None else ""

    def __getitem__(self, key):
        if isinstance(key, slice):
            # Slicing
            eltType = None
            r = []
            for i in range(*key.indices(len(self))):
                if eltType is None:
                    eltType = type(self[i])
                elif not isinstance(self[i], eltType):
                    eltType = Element  # Fall back to element
                r.append(self[i])
            return self.__create_array("", eltType, r)

        elif isinstance(key, str):
            fields = key.split(":")

            if len(fields) <= 1:
                # Selection by name
                eltType = None
                r = []
                for e in self:
                    if fnmatch.fnmatch(e.get_name(), key):
                        if eltType is None:
                            eltType = type(e)
                        elif not isinstance(e, eltType):
                            eltType = Element  # Fall back to element
                        r.append(e)
            else:
                # Selection by fields
                eltType = None
                r = []
                for e in self:
                    txt = self.__eval_field(fields[0], e)
                    if fnmatch.fnmatch(txt, fields[1]):
                        if eltType is None:
                            eltType = type(e)
                        elif not isinstance(e, eltType):
                            eltType = Element  # Fall back to element
                        r.append(e)

            return self.__create_array("", eltType, r)

        else:
            # Default to super selection
            return super().__getitem__(key)
